#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
DRAC inspection interface
"""

from ironic_lib import metrics_utils
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import units

from ironic.common import boot_modes
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import states
from ironic.common import utils
from ironic.drivers import base
from ironic.drivers.modules.drac import common as drac_common
from ironic.drivers.modules.drac import utils as drac_utils
from ironic.drivers.modules import inspect_utils
from ironic.drivers.modules.redfish import inspect as redfish_inspect
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic import objects

drac_exceptions = importutils.try_import('dracclient.exceptions')
sushy = importutils.try_import('sushy')

LOG = logging.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)

_PXE_DEV_ENABLED_INTERFACES = [('PxeDev1EnDis', 'PxeDev1Interface'),
                               ('PxeDev2EnDis', 'PxeDev2Interface'),
                               ('PxeDev3EnDis', 'PxeDev3Interface'),
                               ('PxeDev4EnDis', 'PxeDev4Interface')]
_BIOS_ENABLED_VALUE = 'Enabled'


class DracRedfishInspect(redfish_inspect.RedfishInspect):
    """iDRAC Redfish interface for inspection-related actions."""

    def inspect_hardware(self, task):
        """Inspect hardware to get the hardware properties.

        Inspects hardware to get the essential properties.
        It fails if any of the essential properties
        are not received from the node.

        :param task: a TaskManager instance.
        :raises: HardwareInspectionFailure if essential properties
                 could not be retrieved successfully.
        :returns: The resulting state of inspection.

        """
        # Ensure we create a port for every NIC port found for consistency
        # with our WSMAN inspect behavior and to work around a bug in some
        # versions of the firmware where the port state is not being
        # reported correctly.

        ethernet_interfaces_mac = list(self._get_mac_address(task).values())
        inspect_utils.create_ports_if_not_exist(task, ethernet_interfaces_mac)
        return super(DracRedfishInspect, self).inspect_hardware(task)

    def _get_mac_address(self, task):
        """Get a list of MAC addresses

        :param task: a TaskManager instance.
        :returns: a mapping of interface identities to MAC addresses.
        """
        system = redfish_utils.get_system(task.node)
        # Get dictionary of ethernet interfaces
        if system.ethernet_interfaces and system.ethernet_interfaces.summary:
            ethernet_interfaces = system.ethernet_interfaces.get_members()
            ethernet_interfaces_mac = {
                interface.identity: interface.mac_address
                for interface in ethernet_interfaces}
            return ethernet_interfaces_mac
        else:
            return {}

    def _get_pxe_port_macs(self, task):
        """Get a list of PXE port MAC addresses.

        :param task: a TaskManager instance.
        :returns: Returns list of PXE port MAC addresses.
        """
        system = redfish_utils.get_system(task.node)
        ethernet_interfaces_mac = self._get_mac_address(task)
        pxe_port_macs = []

        if system.boot.mode == boot_modes.UEFI:
            # When a server is in UEFI boot mode, the PXE NIC ports are
            # stored in the PxeDevXEnDis and PxeDevXInterface BIOS
            # settings. Get the PXE NIC ports from these settings and
            # their MAC addresses.
            for param, nic in _PXE_DEV_ENABLED_INTERFACES:
                if system.bios.attributes[param] == _BIOS_ENABLED_VALUE:
                    nic_id = system.bios.attributes[nic]
                    # Get MAC address of the given nic_id
                    mac_address = ethernet_interfaces_mac[nic_id]
                    pxe_port_macs.append(mac_address)
        elif system.boot.mode == boot_modes.LEGACY_BIOS:
            # When a server is in BIOS boot mode, whether or not a
            # NIC port is set to PXE boot is stored on the NIC port
            # itself internally to the BMC. Getting this information
            # requires using an OEM extension to export the system
            # configuration, as the redfish standard does not specify
            # how to get it, and Dell does not have OEM redfish calls
            # to selectively retrieve it at this time.
            # Get instance of Sushy OEM manager object
            pxe_port_macs_list = drac_utils.execute_oem_manager_method(
                task, 'get PXE port MAC addresses',
                lambda m: m.get_pxe_port_macs_bios(ethernet_interfaces_mac))
            pxe_port_macs = [mac for mac in pxe_port_macs_list]

        return pxe_port_macs


class DracWSManInspect(base.InspectInterface):

    _GPU_SUPPORTED_LIST = {"TU104GL [Tesla T4]",
                           "GV100GL [Tesla V100 PCIe 16GB]"}

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return drac_common.COMMON_PROPERTIES

    @METRICS.timer('DracInspect.validate')
    def validate(self, task):
        """Validate the driver-specific info supplied.

        This method validates whether the 'driver_info' property of the
        supplied node contains the required information for this driver to
        manage the node.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue if required driver_info attribute
                 is missing or invalid on the node.

        """
        return drac_common.parse_driver_info(task.node)

    @METRICS.timer('DracInspect.inspect_hardware')
    def inspect_hardware(self, task):
        """Inspect hardware.

        Inspect hardware to obtain the essential & additional hardware
        properties.

        :param task: a TaskManager instance containing the node to act on.
        :raises: HardwareInspectionFailure, if unable to get essential
                 hardware properties.
        :returns: states.MANAGEABLE
        """

        node = task.node
        client = drac_common.get_drac_client(node)
        properties = {}

        try:
            properties['memory_mb'] = sum(
                [memory.size_mb for memory in client.list_memory()])
            cpus = client.list_cpus()
            if cpus:
                properties['cpus'] = sum(
                    [self._calculate_cpus(cpu) for cpu in cpus])
                properties['cpu_arch'] = 'x86_64' if cpus[0].arch64 else 'x86'

            bios_settings = client.list_bios_settings()
            video_controllers = client.list_video_controllers()
            current_capabilities = node.properties.get('capabilities', '')
            new_capabilities = {
                'boot_mode': bios_settings["BootMode"].current_value.lower(),
                'pci_gpu_devices': self._calculate_gpus(video_controllers)}

            capabilties = utils.get_updated_capabilities(current_capabilities,
                                                         new_capabilities)
            properties['capabilities'] = capabilties

            virtual_disks = client.list_virtual_disks()
            root_disk = self._guess_root_disk(virtual_disks)
            if root_disk:
                properties['local_gb'] = int(root_disk.size_mb / units.Ki)
            else:
                physical_disks = client.list_physical_disks()
                root_disk = self._guess_root_disk(physical_disks)
                if root_disk:
                    properties['local_gb'] = int(
                        root_disk.size_mb / units.Ki)
        except drac_exceptions.BaseClientException as exc:
            LOG.error('DRAC driver failed to introspect node '
                      '%(node_uuid)s. Reason: %(error)s.',
                      {'node_uuid': node.uuid, 'error': exc})
            raise exception.HardwareInspectionFailure(error=exc)

        valid_keys = self.ESSENTIAL_PROPERTIES
        missing_keys = valid_keys - set(properties)
        if missing_keys:
            error = (_('Failed to discover the following properties: '
                       '%(missing_keys)s') %
                     {'missing_keys': ', '.join(missing_keys)})
            raise exception.HardwareInspectionFailure(error=error)

        node.properties = dict(node.properties, **properties)
        node.save()

        try:
            nics = client.list_nics()
        except drac_exceptions.BaseClientException as exc:
            LOG.error('DRAC driver failed to introspect node '
                      '%(node_uuid)s. Reason: %(error)s.',
                      {'node_uuid': node.uuid, 'error': exc})
            raise exception.HardwareInspectionFailure(error=exc)

        pxe_dev_nics = self._get_pxe_dev_nics(client, nics, node)
        if pxe_dev_nics is None:
            LOG.warning('No PXE enabled NIC was found for node '
                        '%(node_uuid)s.', {'node_uuid': node.uuid})

        for nic in nics:
            try:
                port = objects.Port(task.context, address=nic.mac,
                                    node_id=node.id,
                                    pxe_enabled=(nic.id in pxe_dev_nics))
                port.create()

                LOG.info('Port created with MAC address %(mac)s '
                         'for node %(node_uuid)s during inspection',
                         {'mac': nic.mac, 'node_uuid': node.uuid})
            except exception.MACAlreadyExists:
                LOG.warning('Failed to create a port with MAC address '
                            '%(mac)s when inspecting the node '
                            '%(node_uuid)s because the address is already '
                            'registered',
                            {'mac': nic.mac, 'node_uuid': node.uuid})

        LOG.info('Node %s successfully inspected.', node.uuid)
        return states.MANAGEABLE

    def _guess_root_disk(self, disks, min_size_required_mb=4 * units.Ki):
        """Find a root disk.

        :param disks: list of disks.
        :param min_size_required_mb: minimum required size of the root disk in
                                     megabytes.
        :returns: root disk.
        """
        disks.sort(key=lambda disk: disk.size_mb)
        for disk in disks:
            if disk.size_mb >= min_size_required_mb:
                return disk

    def _calculate_cpus(self, cpu):
        """Find actual CPU count.

        :param cpu: Pass cpu.

        :returns: returns total cpu count.
        """
        if cpu.ht_enabled:
            return cpu.cores * 2
        else:
            return cpu.cores

    def _calculate_gpus(self, video_controllers):
        """Find actual GPU count.

        This method reports number of NVIDIA Tesla T4 GPU devices present
        on the server.

        :param video_controllers: list of video controllers.

        :returns: returns total gpu count.
        """
        gpu_cnt = 0
        for video_controller in video_controllers:
            for gpu in self._GPU_SUPPORTED_LIST:
                if video_controller.description == gpu:
                    gpu_cnt += 1
        return gpu_cnt

    def _get_pxe_dev_nics(self, client, nics, node):
        """Get a list of pxe device interfaces.

        :param client: Dracclient to list the bios settings and nics
        :param nics: list of nics

        :returns: Returns list of pxe device interfaces.
        """
        pxe_dev_nics = []
        pxe_params = ["PxeDev1EnDis", "PxeDev2EnDis",
                      "PxeDev3EnDis", "PxeDev4EnDis"]
        pxe_nics = ["PxeDev1Interface", "PxeDev2Interface",
                    "PxeDev3Interface", "PxeDev4Interface"]

        try:
            bios_settings = client.list_bios_settings()
        except drac_exceptions.BaseClientException as exc:
            LOG.error('DRAC driver failed to list bios settings '
                      'for %(node_uuid)s. Reason: %(error)s.',
                      {'node_uuid': node.uuid, 'error': exc})
            raise exception.HardwareInspectionFailure(error=exc)

        if bios_settings["BootMode"].current_value == "Uefi":
            for param, nic in zip(pxe_params, pxe_nics):
                if param in bios_settings and bios_settings[
                        param].current_value == "Enabled":
                    pxe_dev_nics.append(
                        bios_settings[nic].current_value)
        elif bios_settings["BootMode"].current_value == "Bios":
            for nic in nics:
                try:
                    nic_cap = client.list_nic_settings(nic_id=nic.id)
                except drac_exceptions.BaseClientException as exc:
                    LOG.error('DRAC driver failed to list nic settings '
                              'for %(node_uuid)s. Reason: %(error)s.',
                              {'node_uuid': node.uuid, 'error': exc})
                    raise exception.HardwareInspectionFailure(error=exc)

                if ("LegacyBootProto" in nic_cap and nic_cap[
                        'LegacyBootProto'].current_value == "PXE"):
                    pxe_dev_nics.append(nic.id)

        return pxe_dev_nics


class DracInspect(DracWSManInspect):
    """Class alias of class DracWSManInspect.

    This class provides ongoing support of the deprecated 'idrac'
    inspect interface implementation entrypoint.

    All bug fixes and new features should be implemented in its base
    class, DracWSManInspect. That makes them available to both the
    deprecated 'idrac' and new 'idrac-wsman' entrypoints. Such changes
    should not be made to this class.
    """

    def __init__(self):
        super(DracInspect, self).__init__()
        LOG.warning("Inspect interface 'idrac' is deprecated and may be "
                    "removed in a future release. Use 'idrac-wsman' instead.")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic.common import exception
from ironic.db.sqlalchemy import models
from ironic.tests import base as test_base


class TestGetClass(test_base.TestCase):

    def test_get_class(self):
        ret = models.get_class('Chassis')
        self.assertEqual(models.Chassis, ret)

        for model in models.Base.__subclasses__():
            ret = models.get_class(model.__name__)
            self.assertEqual(model, ret)

    def test_get_class_bad(self):
        self.assertRaises(exception.IronicException,
                          models.get_class, "DoNotExist")



# Copyright 2014 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from ironic.common import components
from ironic.common import exception
from ironic.common import indicator_states
from ironic.common import raid
from ironic.common import states
from ironic.drivers import base as driver_base
from ironic.drivers.modules import fake
from ironic.tests import base


class FakeVendorInterface(driver_base.VendorInterface):
    def get_properties(self):
        pass

    @driver_base.passthru(['POST'])
    def noexception(self):
        return "Fake"

    @driver_base.driver_passthru(['POST'])
    def driver_noexception(self):
        return "Fake"

    @driver_base.passthru(['POST'])
    def ironicexception(self):
        raise exception.IronicException("Fake!")

    @driver_base.passthru(['POST'])
    def normalexception(self):
        raise Exception("Fake!")

    @driver_base.passthru(['POST'], require_exclusive_lock=False)
    def shared_task(self):
        return "shared fake"

    def validate(self, task, **kwargs):
        pass

    def driver_validate(self, **kwargs):
        pass


class PassthruDecoratorTestCase(base.TestCase):

    def setUp(self):
        super(PassthruDecoratorTestCase, self).setUp()
        self.fvi = FakeVendorInterface()

    def test_passthru_noexception(self):
        result = self.fvi.noexception()
        self.assertEqual("Fake", result)

    @mock.patch.object(driver_base, 'LOG', autospec=True)
    def test_passthru_ironicexception(self, mock_log):
        self.assertRaises(exception.IronicException,
                          self.fvi.ironicexception, mock.ANY)
        mock_log.exception.assert_called_with(
            mock.ANY, 'ironicexception')

    @mock.patch.object(driver_base, 'LOG', autospec=True)
    def test_passthru_nonironicexception(self, mock_log):
        self.assertRaises(exception.VendorPassthruException,
                          self.fvi.normalexception, mock.ANY)
        mock_log.exception.assert_called_with(
            mock.ANY, 'normalexception')

    def test_passthru_shared_task_metadata(self):
        self.assertIn('require_exclusive_lock',
                      self.fvi.shared_task._vendor_metadata[1])
        self.assertFalse(
            self.fvi.shared_task._vendor_metadata[1]['require_exclusive_lock'])

    def test_passthru_exclusive_task_metadata(self):
        self.assertIn('require_exclusive_lock',
                      self.fvi.noexception._vendor_metadata[1])
        self.assertTrue(
            self.fvi.noexception._vendor_metadata[1]['require_exclusive_lock'])

    def test_passthru_check_func_references(self):
        inst1 = FakeVendorInterface()
        inst2 = FakeVendorInterface()

        self.assertNotEqual(inst1.vendor_routes['noexception']['func'],
                            inst2.vendor_routes['noexception']['func'])
        self.assertNotEqual(inst1.driver_routes['driver_noexception']['func'],
                            inst2.driver_routes['driver_noexception']['func'])


class CleanStepDecoratorTestCase(base.TestCase):

    def setUp(self):
        super(CleanStepDecoratorTestCase, self).setUp()
        method_mock = mock.MagicMock()
        del method_mock._is_clean_step
        del method_mock._clean_step_priority
        del method_mock._clean_step_abortable
        del method_mock._clean_step_argsinfo
        self.method = method_mock

    def test__validate_argsinfo(self):
        # None, empty dict
        driver_base._validate_argsinfo(None)
        driver_base._validate_argsinfo({})

        # Only description specified
        driver_base._validate_argsinfo({'arg1': {'description': 'desc1'}})

        # Multiple args
        driver_base._validate_argsinfo({'arg1': {'description': 'desc1',
                                                 'required': True},
                                        'arg2': {'description': 'desc2'}})

    def test__validate_argsinfo_not_dict(self):
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'argsinfo.+dictionary',
                               driver_base._validate_argsinfo, 'not-a-dict')

    def test__validate_argsinfo_arg_not_dict(self):
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'Argument.+dictionary',
                               driver_base._validate_argsinfo,
                               {'arg1': 'not-a-dict'})

    def test__validate_argsinfo_arg_empty_dict(self):
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'description',
                               driver_base._validate_argsinfo,
                               {'arg1': {}})

    def test__validate_argsinfo_arg_missing_description(self):
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'description',
                               driver_base._validate_argsinfo,
                               {'arg1': {'required': True}})

    def test__validate_argsinfo_arg_description_invalid(self):
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'string',
                               driver_base._validate_argsinfo,
                               {'arg1': {'description': True}})

    def test__validate_argsinfo_arg_required_invalid(self):
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'Boolean',
                               driver_base._validate_argsinfo,
                               {'arg1': {'description': 'desc1',
                                         'required': 'maybe'}})

    def test__validate_argsinfo_arg_unknown_key(self):
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'invalid',
                               driver_base._validate_argsinfo,
                               {'arg1': {'description': 'desc1',
                                         'unknown': 'bad'}})

    def test_clean_step_priority_only(self):
        d = driver_base.clean_step(priority=10)
        d(self.method)
        self.assertTrue(self.method._is_clean_step)
        self.assertEqual(10, self.method._clean_step_priority)
        self.assertFalse(self.method._clean_step_abortable)
        self.assertIsNone(self.method._clean_step_argsinfo)

    def test_clean_step_all_args(self):
        argsinfo = {'arg1': {'description': 'desc1',
                             'required': True}}
        d = driver_base.clean_step(priority=0, abortable=True,
                                   argsinfo=argsinfo)
        d(self.method)
        self.assertTrue(self.method._is_clean_step)
        self.assertEqual(0, self.method._clean_step_priority)
        self.assertTrue(self.method._clean_step_abortable)
        self.assertEqual(argsinfo, self.method._clean_step_argsinfo)

    def test_clean_step_bad_priority(self):
        d = driver_base.clean_step(priority='hi')
        self.assertRaisesRegex(exception.InvalidParameterValue, 'priority',
                               d, self.method)
        self.assertTrue(self.method._is_clean_step)
        self.assertFalse(hasattr(self.method, '_clean_step_priority'))
        self.assertFalse(hasattr(self.method, '_clean_step_abortable'))
        self.assertFalse(hasattr(self.method, '_clean_step_argsinfo'))

    def test_clean_step_bad_abortable(self):
        d = driver_base.clean_step(priority=0, abortable='blue')
        self.assertRaisesRegex(exception.InvalidParameterValue, 'abortable',
                               d, self.method)
        self.assertTrue(self.method._is_clean_step)
        self.assertEqual(0, self.method._clean_step_priority)
        self.assertFalse(hasattr(self.method, '_clean_step_abortable'))
        self.assertFalse(hasattr(self.method, '_clean_step_argsinfo'))

    @mock.patch.object(driver_base, '_validate_argsinfo', spec_set=True,
                       autospec=True)
    def test_clean_step_bad_argsinfo(self, mock_valid):
        mock_valid.side_effect = exception.InvalidParameterValue('bad')
        d = driver_base.clean_step(priority=0, argsinfo=100)
        self.assertRaises(exception.InvalidParameterValue, d, self.method)
        self.assertTrue(self.method._is_clean_step)
        self.assertEqual(0, self.method._clean_step_priority)
        self.assertFalse(self.method._clean_step_abortable)
        self.assertFalse(hasattr(self.method, '_clean_step_argsinfo'))


class CleanStepTestCase(base.TestCase):
    def test_get_and_execute_clean_steps(self):
        # Create a fake Driver class, create some clean steps, make sure
        # they are listed correctly, and attempt to execute one of them

        method_mock = mock.MagicMock(spec_set=[])
        method_args_mock = mock.MagicMock(spec_set=[])
        task_mock = mock.MagicMock(spec_set=[])

        class BaseTestClass(driver_base.BaseInterface):
            def get_properties(self):
                return {}

            def validate(self, task):
                pass

        class TestClass(BaseTestClass):
            interface_type = 'test'

            @driver_base.clean_step(priority=0)
            def manual_method(self, task):
                pass

            @driver_base.clean_step(priority=10, abortable=True)
            def automated_method(self, task):
                method_mock(task)

            def not_clean_method(self, task):
                pass

        class TestClass2(BaseTestClass):
            interface_type = 'test2'

            @driver_base.clean_step(priority=0)
            def manual_method2(self, task):
                pass

            @driver_base.clean_step(priority=20, abortable=True)
            def automated_method2(self, task):
                method_mock(task)

            def not_clean_method2(self, task):
                pass

        class TestClass3(BaseTestClass):
            interface_type = 'test3'

            @driver_base.clean_step(priority=0, abortable=True, argsinfo={
                                    'arg1': {'description': 'desc1',
                                             'required': True}})
            def manual_method3(self, task, **kwargs):
                method_args_mock(task, **kwargs)

            @driver_base.clean_step(priority=15, argsinfo={
                                    'arg10': {'description': 'desc10'}})
            def automated_method3(self, task, **kwargs):
                pass

            def not_clean_method3(self, task):
                pass

        obj = TestClass()
        obj2 = TestClass2()
        obj3 = TestClass3()

        self.assertEqual(2, len(obj.get_clean_steps(task_mock)))
        # Ensure the steps look correct
        self.assertEqual(10, obj.get_clean_steps(task_mock)[0]['priority'])
        self.assertTrue(obj.get_clean_steps(task_mock)[0]['abortable'])
        self.assertEqual('test', obj.get_clean_steps(
            task_mock)[0]['interface'])
        self.assertEqual('automated_method', obj.get_clean_steps(
            task_mock)[0]['step'])
        self.assertEqual(0, obj.get_clean_steps(task_mock)[1]['priority'])
        self.assertFalse(obj.get_clean_steps(task_mock)[1]['abortable'])
        self.assertEqual('test', obj.get_clean_steps(
            task_mock)[1]['interface'])
        self.assertEqual('manual_method', obj.get_clean_steps(
            task_mock)[1]['step'])

        # Ensure the second obj get different clean steps
        self.assertEqual(2, len(obj2.get_clean_steps(task_mock)))
        # Ensure the steps look correct
        self.assertEqual(20, obj2.get_clean_steps(task_mock)[0]['priority'])
        self.assertTrue(obj2.get_clean_steps(task_mock)[0]['abortable'])
        self.assertEqual('test2', obj2.get_clean_steps(
            task_mock)[0]['interface'])
        self.assertEqual('automated_method2', obj2.get_clean_steps(
            task_mock)[0]['step'])
        self.assertEqual(0, obj2.get_clean_steps(task_mock)[1]['priority'])
        self.assertFalse(obj2.get_clean_steps(task_mock)[1]['abortable'])
        self.assertEqual('test2', obj2.get_clean_steps(
            task_mock)[1]['interface'])
        self.assertEqual('manual_method2', obj2.get_clean_steps(
            task_mock)[1]['step'])
        self.assertIsNone(obj2.get_clean_steps(task_mock)[0]['argsinfo'])

        # Ensure the third obj has different clean steps
        self.assertEqual(2, len(obj3.get_clean_steps(task_mock)))
        self.assertEqual(15, obj3.get_clean_steps(task_mock)[0]['priority'])
        self.assertFalse(obj3.get_clean_steps(task_mock)[0]['abortable'])
        self.assertEqual('test3', obj3.get_clean_steps(
            task_mock)[0]['interface'])
        self.assertEqual('automated_method3', obj3.get_clean_steps(
            task_mock)[0]['step'])
        self.assertEqual({'arg10': {'description': 'desc10'}},
                         obj3.get_clean_steps(task_mock)[0]['argsinfo'])
        self.assertEqual(0, obj3.get_clean_steps(task_mock)[1]['priority'])
        self.assertTrue(obj3.get_clean_steps(task_mock)[1]['abortable'])
        self.assertEqual(obj3.interface_type, obj3.get_clean_steps(
            task_mock)[1]['interface'])
        self.assertEqual('manual_method3', obj3.get_clean_steps(
            task_mock)[1]['step'])
        self.assertEqual({'arg1': {'description': 'desc1', 'required': True}},
                         obj3.get_clean_steps(task_mock)[1]['argsinfo'])

        # Ensure we can execute the function.
        obj.execute_clean_step(task_mock, obj.get_clean_steps(task_mock)[0])
        method_mock.assert_called_once_with(task_mock)

        args = {'arg1': 'val1'}
        clean_step = {'interface': 'test3', 'step': 'manual_method3',
                      'args': args}
        obj3.execute_clean_step(task_mock, clean_step)
        method_args_mock.assert_called_once_with(task_mock, **args)


class DeployStepDecoratorTestCase(base.TestCase):

    def setUp(self):
        super(DeployStepDecoratorTestCase, self).setUp()
        method_mock = mock.MagicMock()
        del method_mock._is_deploy_step
        del method_mock._deploy_step_priority
        del method_mock._deploy_step_argsinfo
        self.method = method_mock

    def test_deploy_step_priority_only(self):
        d = driver_base.deploy_step(priority=10)
        d(self.method)
        self.assertTrue(self.method._is_deploy_step)
        self.assertEqual(10, self.method._deploy_step_priority)
        self.assertIsNone(self.method._deploy_step_argsinfo)

    def test_deploy_step_all_args(self):
        argsinfo = {'arg1': {'description': 'desc1',
                             'required': True}}
        d = driver_base.deploy_step(priority=0, argsinfo=argsinfo)
        d(self.method)
        self.assertTrue(self.method._is_deploy_step)
        self.assertEqual(0, self.method._deploy_step_priority)
        self.assertEqual(argsinfo, self.method._deploy_step_argsinfo)

    def test_deploy_step_bad_priority(self):
        d = driver_base.deploy_step(priority='hi')
        self.assertRaisesRegex(exception.InvalidParameterValue, 'priority',
                               d, self.method)
        self.assertTrue(self.method._is_deploy_step)
        self.assertFalse(hasattr(self.method, '_deploy_step_priority'))
        self.assertFalse(hasattr(self.method, '_deploy_step_argsinfo'))

    @mock.patch.object(driver_base, '_validate_argsinfo', spec_set=True,
                       autospec=True)
    def test_deploy_step_bad_argsinfo(self, mock_valid):
        mock_valid.side_effect = exception.InvalidParameterValue('bad')
        d = driver_base.deploy_step(priority=0, argsinfo=100)
        self.assertRaises(exception.InvalidParameterValue, d, self.method)
        self.assertTrue(self.method._is_deploy_step)
        self.assertEqual(0, self.method._deploy_step_priority)
        self.assertFalse(hasattr(self.method, '_deploy_step_argsinfo'))


class DeployAndCleanStepDecoratorTestCase(base.TestCase):

    def setUp(self):
        super(DeployAndCleanStepDecoratorTestCase, self).setUp()
        method_mock = mock.MagicMock()
        del method_mock._is_deploy_step
        del method_mock._deploy_step_priority
        del method_mock._deploy_step_argsinfo
        del method_mock._is_clean_step
        del method_mock._clean_step_priority
        del method_mock._clean_step_abortable
        del method_mock._clean_step_argsinfo
        self.method = method_mock

    def test_deploy_and_clean_step_priority_only(self):
        dd = driver_base.deploy_step(priority=10)
        dc = driver_base.clean_step(priority=11)
        dd(dc(self.method))
        self.assertTrue(self.method._is_deploy_step)
        self.assertEqual(10, self.method._deploy_step_priority)
        self.assertIsNone(self.method._deploy_step_argsinfo)
        self.assertTrue(self.method._is_clean_step)
        self.assertEqual(11, self.method._clean_step_priority)
        self.assertFalse(self.method._clean_step_abortable)
        self.assertIsNone(self.method._clean_step_argsinfo)

    def test_deploy_and_clean_step_all_args(self):
        dargsinfo = {'arg1': {'description': 'desc1',
                              'required': True}}
        cargsinfo = {'arg2': {'description': 'desc2',
                              'required': False}}
        dd = driver_base.deploy_step(priority=0, argsinfo=dargsinfo)
        dc = driver_base.clean_step(priority=0, argsinfo=cargsinfo)
        dd(dc(self.method))
        self.assertTrue(self.method._is_deploy_step)
        self.assertEqual(0, self.method._deploy_step_priority)
        self.assertEqual(dargsinfo, self.method._deploy_step_argsinfo)
        self.assertTrue(self.method._is_clean_step)
        self.assertEqual(0, self.method._clean_step_priority)
        self.assertFalse(self.method._clean_step_abortable)
        self.assertEqual(cargsinfo, self.method._clean_step_argsinfo)

    def test_clean_and_deploy_step_all_args(self):
        # Opposite ordering, should make no difference.
        dargsinfo = {'arg1': {'description': 'desc1',
                              'required': True}}
        cargsinfo = {'arg2': {'description': 'desc2',
                              'required': False}}
        dd = driver_base.deploy_step(priority=0, argsinfo=dargsinfo)
        dc = driver_base.clean_step(priority=0, argsinfo=cargsinfo)
        dc(dd(self.method))
        self.assertTrue(self.method._is_deploy_step)
        self.assertEqual(0, self.method._deploy_step_priority)
        self.assertEqual(dargsinfo, self.method._deploy_step_argsinfo)
        self.assertTrue(self.method._is_clean_step)
        self.assertEqual(0, self.method._clean_step_priority)
        self.assertFalse(self.method._clean_step_abortable)
        self.assertEqual(cargsinfo, self.method._clean_step_argsinfo)


class DeployStepTestCase(base.TestCase):
    def test_get_and_execute_deploy_steps(self):
        # Create a fake Driver class, create some deploy steps, make sure
        # they are listed correctly, and attempt to execute one of them

        method_mock = mock.MagicMock(spec_set=[])
        method_args_mock = mock.MagicMock(spec_set=[])
        task_mock = mock.MagicMock(spec_set=[])

        class BaseTestClass(driver_base.BaseInterface):
            def get_properties(self):
                return {}

            def validate(self, task):
                pass

        class TestClass(BaseTestClass):
            interface_type = 'test'

            @driver_base.deploy_step(priority=0)
            def deploy_zero(self, task):
                pass

            @driver_base.deploy_step(priority=10)
            def deploy_ten(self, task):
                method_mock(task)

            def not_deploy_method(self, task):
                pass

        class TestClass2(BaseTestClass):
            interface_type = 'test2'

            @driver_base.deploy_step(priority=0)
            def deploy_zero2(self, task):
                pass

            @driver_base.deploy_step(priority=20)
            def deploy_twenty(self, task):
                method_mock(task)

            def not_deploy_method2(self, task):
                pass

        class TestClass3(BaseTestClass):
            interface_type = 'test3'

            @driver_base.deploy_step(priority=0, argsinfo={
                                     'arg1': {'description': 'desc1',
                                              'required': True}})
            def deploy_zero3(self, task, **kwargs):
                method_args_mock(task, **kwargs)

            @driver_base.deploy_step(priority=15, argsinfo={
                                     'arg10': {'description': 'desc10'}})
            def deploy_fifteen(self, task, **kwargs):
                pass

            def not_deploy_method3(self, task):
                pass

        obj = TestClass()
        obj2 = TestClass2()
        obj3 = TestClass3()

        self.assertEqual(2, len(obj.get_deploy_steps(task_mock)))
        # Ensure the steps look correct
        self.assertEqual(10, obj.get_deploy_steps(task_mock)[0]['priority'])
        self.assertEqual('test', obj.get_deploy_steps(
            task_mock)[0]['interface'])
        self.assertEqual('deploy_ten', obj.get_deploy_steps(
            task_mock)[0]['step'])
        self.assertEqual(0, obj.get_deploy_steps(task_mock)[1]['priority'])
        self.assertEqual('test', obj.get_deploy_steps(
            task_mock)[1]['interface'])
        self.assertEqual('deploy_zero', obj.get_deploy_steps(
            task_mock)[1]['step'])

        # Ensure the second obj has different deploy steps
        self.assertEqual(2, len(obj2.get_deploy_steps(task_mock)))
        # Ensure the steps look correct
        self.assertEqual(20, obj2.get_deploy_steps(task_mock)[0]['priority'])
        self.assertEqual('test2', obj2.get_deploy_steps(
            task_mock)[0]['interface'])
        self.assertEqual('deploy_twenty', obj2.get_deploy_steps(
            task_mock)[0]['step'])
        self.assertEqual(0, obj2.get_deploy_steps(task_mock)[1]['priority'])
        self.assertEqual('test2', obj2.get_deploy_steps(
            task_mock)[1]['interface'])
        self.assertEqual('deploy_zero2', obj2.get_deploy_steps(
            task_mock)[1]['step'])
        self.assertIsNone(obj2.get_deploy_steps(task_mock)[0]['argsinfo'])

        # Ensure the third obj has different deploy steps
        self.assertEqual(2, len(obj3.get_deploy_steps(task_mock)))
        self.assertEqual(15, obj3.get_deploy_steps(task_mock)[0]['priority'])
        self.assertEqual('test3', obj3.get_deploy_steps(
            task_mock)[0]['interface'])
        self.assertEqual('deploy_fifteen', obj3.get_deploy_steps(
            task_mock)[0]['step'])
        self.assertEqual({'arg10': {'description': 'desc10'}},
                         obj3.get_deploy_steps(task_mock)[0]['argsinfo'])
        self.assertEqual(0, obj3.get_deploy_steps(task_mock)[1]['priority'])
        self.assertEqual(obj3.interface_type, obj3.get_deploy_steps(
            task_mock)[1]['interface'])
        self.assertEqual('deploy_zero3', obj3.get_deploy_steps(
            task_mock)[1]['step'])
        self.assertEqual({'arg1': {'description': 'desc1', 'required': True}},
                         obj3.get_deploy_steps(task_mock)[1]['argsinfo'])

        # Ensure we can execute the function.
        obj.execute_deploy_step(task_mock, obj.get_deploy_steps(task_mock)[0])
        method_mock.assert_called_once_with(task_mock)

        args = {'arg1': 'val1'}
        deploy_step = {'interface': 'test3', 'step': 'deploy_zero3',
                       'args': args}
        obj3.execute_deploy_step(task_mock, deploy_step)
        method_args_mock.assert_called_once_with(task_mock, **args)


class MyRAIDInterface(driver_base.RAIDInterface):

    def create_configuration(self, task,
                             create_root_volume=True,
                             create_nonroot_volumes=True,
                             delete_existing=True):
        pass

    def delete_configuration(self, task):
        pass


class RAIDInterfaceTestCase(base.TestCase):

    @mock.patch.object(driver_base.RAIDInterface, 'validate_raid_config',
                       autospec=True)
    def test_validate(self, validate_raid_config_mock):
        raid_interface = MyRAIDInterface()
        node_mock = mock.MagicMock(target_raid_config='some_raid_config')
        task_mock = mock.MagicMock(node=node_mock)

        raid_interface.validate(task_mock)

        validate_raid_config_mock.assert_called_once_with(
            raid_interface, task_mock, 'some_raid_config')

    @mock.patch.object(driver_base.RAIDInterface, 'validate_raid_config',
                       autospec=True)
    def test_validate_no_target_raid_config(self, validate_raid_config_mock):
        raid_interface = MyRAIDInterface()
        node_mock = mock.MagicMock(target_raid_config={})
        task_mock = mock.MagicMock(node=node_mock)

        raid_interface.validate(task_mock)

        self.assertFalse(validate_raid_config_mock.called)

    @mock.patch.object(raid, 'validate_configuration', autospec=True)
    def test_validate_raid_config(self, common_validate_mock):
        with open(driver_base.RAID_CONFIG_SCHEMA, 'r') as raid_schema_fobj:
            raid_schema = json.load(raid_schema_fobj)
        raid_interface = MyRAIDInterface()

        raid_interface.validate_raid_config('task', 'some_raid_config')

        common_validate_mock.assert_called_once_with(
            'some_raid_config', raid_schema)

    @mock.patch.object(raid, 'get_logical_disk_properties',
                       autospec=True)
    def test_get_logical_disk_properties(self, get_properties_mock):
        with open(driver_base.RAID_CONFIG_SCHEMA, 'r') as raid_schema_fobj:
            raid_schema = json.load(raid_schema_fobj)
        raid_interface = MyRAIDInterface()
        raid_interface.get_logical_disk_properties()
        get_properties_mock.assert_called_once_with(raid_schema)

    @mock.patch.object(MyRAIDInterface, 'create_configuration', autospec=True)
    @mock.patch.object(MyRAIDInterface, 'validate_raid_config',
                       autospec=True)
    def test_apply_configuration(self, mock_validate, mock_create):
        raid_interface = MyRAIDInterface()
        node_mock = mock.MagicMock(target_raid_config=None)
        task_mock = mock.MagicMock(node=node_mock)
        mock_create.return_value = states.DEPLOYWAIT
        raid_config = 'some_raid_config'

        result = raid_interface.apply_configuration(task_mock, raid_config)

        self.assertEqual(states.DEPLOYWAIT, result)
        mock_validate.assert_called_once_with(raid_interface, task_mock,
                                              raid_config)
        mock_create.assert_called_once_with(raid_interface, task_mock,
                                            create_root_volume=True,
                                            create_nonroot_volumes=True,
                                            delete_existing=True)
        self.assertEqual(raid_config, node_mock.target_raid_config)

    @mock.patch.object(MyRAIDInterface, 'create_configuration', autospec=True)
    @mock.patch.object(MyRAIDInterface, 'validate_raid_config',
                       autospec=True)
    def test_apply_configuration_delete_existing(self, mock_validate,
                                                 mock_create):
        raid_interface = MyRAIDInterface()
        node_mock = mock.MagicMock(target_raid_config=None)
        task_mock = mock.MagicMock(node=node_mock)
        mock_create.return_value = states.DEPLOYWAIT
        raid_config = 'some_raid_config'

        result = raid_interface.apply_configuration(task_mock, raid_config,
                                                    delete_existing=True)

        self.assertEqual(states.DEPLOYWAIT, result)
        mock_validate.assert_called_once_with(raid_interface, task_mock,
                                              raid_config)
        mock_create.assert_called_once_with(raid_interface, task_mock,
                                            create_root_volume=True,
                                            create_nonroot_volumes=True,
                                            delete_existing=True)
        self.assertEqual(raid_config, node_mock.target_raid_config)

    @mock.patch.object(MyRAIDInterface, 'create_configuration', autospec=True)
    @mock.patch.object(MyRAIDInterface, 'validate_raid_config',
                       autospec=True)
    def test_apply_configuration_invalid(self, mock_validate, mock_create):
        raid_interface = MyRAIDInterface()
        node_mock = mock.MagicMock(target_raid_config=None)
        task_mock = mock.MagicMock(node=node_mock)
        mock_validate.side_effect = exception.InvalidParameterValue('bad')
        raid_config = 'some_raid_config'

        self.assertRaises(exception.InvalidParameterValue,
                          raid_interface.apply_configuration, task_mock,
                          raid_config)
        mock_validate.assert_called_once_with(raid_interface, task_mock,
                                              raid_config)
        self.assertFalse(mock_create.called)
        self.assertIsNone(node_mock.target_raid_config)


class TestDeployInterface(base.TestCase):
    @mock.patch.object(driver_base.LOG, 'warning', autospec=True)
    def test_warning_on_heartbeat(self, mock_log):
        # NOTE(dtantsur): FakeDeploy does not override heartbeat
        deploy = fake.FakeDeploy()
        deploy.heartbeat(mock.Mock(node=mock.Mock(uuid='uuid',
                                                  driver='driver')),
                         'url', '3.2.0')
        self.assertTrue(mock_log.called)


class MyBIOSInterface(driver_base.BIOSInterface):

    def get_properties(self):
        pass

    def validate(self, task):
        pass

    @driver_base.cache_bios_settings
    def apply_configuration(self, task, settings):
        return "return_value_apply_configuration"

    @driver_base.cache_bios_settings
    def factory_reset(self, task):
        return "return_value_factory_reset"

    def cache_bios_settings(self, task):
        pass


class TestBIOSInterface(base.TestCase):

    @mock.patch.object(MyBIOSInterface, 'cache_bios_settings', autospec=True)
    def test_apply_configuration_wrapper(self, cache_bios_settings_mock):
        bios = MyBIOSInterface()
        task_mock = mock.MagicMock()

        actual = bios.apply_configuration(task_mock, "")
        cache_bios_settings_mock.assert_called_once_with(bios, task_mock)
        self.assertEqual(actual, "return_value_apply_configuration")

    @mock.patch.object(MyBIOSInterface, 'cache_bios_settings', autospec=True)
    def test_factory_reset_wrapper(self, cache_bios_settings_mock):
        bios = MyBIOSInterface()
        task_mock = mock.MagicMock()

        actual = bios.factory_reset(task_mock)
        cache_bios_settings_mock.assert_called_once_with(bios, task_mock)
        self.assertEqual(actual, "return_value_factory_reset")


class TestBootInterface(base.TestCase):

    def test_validate_rescue_default_impl(self):
        boot = fake.FakeBoot()
        task_mock = mock.MagicMock(spec_set=['node'])

        self.assertRaises(exception.UnsupportedDriverExtension,
                          boot.validate_rescue, task_mock)


class TestManagementInterface(base.TestCase):

    def test_inject_nmi_default_impl(self):
        management = fake.FakeManagement()
        task_mock = mock.MagicMock(spec_set=['node'])

        self.assertRaises(exception.UnsupportedDriverExtension,
                          management.inject_nmi, task_mock)

    def test_get_supported_boot_modes_default_impl(self):
        management = fake.FakeManagement()
        task_mock = mock.MagicMock(spec_set=['node'])

        self.assertRaises(exception.UnsupportedDriverExtension,
                          management.get_supported_boot_modes, task_mock)

    def test_set_boot_mode_default_impl(self):
        management = fake.FakeManagement()
        task_mock = mock.MagicMock(spec_set=['node'])

        self.assertRaises(exception.UnsupportedDriverExtension,
                          management.set_boot_mode, task_mock, 'whatever')

    def test_get_boot_mode_default_impl(self):
        management = fake.FakeManagement()
        task_mock = mock.MagicMock(spec_set=['node'])

        self.assertRaises(exception.UnsupportedDriverExtension,
                          management.get_boot_mode, task_mock)

    def test_get_supported_indicators_default_impl(self):
        management = fake.FakeManagement()
        task_mock = mock.MagicMock(spec_set=['node'])

        expected = {
            components.CHASSIS: {
                'led-0': {
                    "readonly": True,
                    "states": [
                        indicator_states.OFF,
                        indicator_states.ON
                    ]
                }
            },
            components.SYSTEM: {
                'led': {
                    "readonly": False,
                    "states": [
                        indicator_states.BLINKING,
                        indicator_states.OFF,
                        indicator_states.ON
                    ]
                }
            }
        }

        self.assertEqual(
            expected, management.get_supported_indicators(task_mock))

    def test_set_indicator_state_default_impl(self):
        management = fake.FakeManagement()
        task_mock = mock.MagicMock(spec_set=['node'])

        self.assertRaises(exception.UnsupportedDriverExtension,
                          management.set_indicator_state, task_mock,
                          components.CHASSIS, 'led-0', indicator_states.ON)

    def test_get_indicator_state_default_impl(self):
        management = fake.FakeManagement()
        task_mock = mock.MagicMock(spec_set=['node'])

        expected = indicator_states.ON

        self.assertEqual(
            expected, management.get_indicator_state(
                task_mock, components.CHASSIS, 'led-0'))

    def test_get_mac_addresses(self):
        management = fake.FakeManagement()
        task_mock = mock.MagicMock(spec_set=['node'])

        self.assertRaises(exception.UnsupportedDriverExtension,
                          management.get_mac_addresses, task_mock)


class TestBareDriver(base.TestCase):

    def test_class_variables(self):
        self.assertEqual(['boot', 'deploy', 'management', 'network', 'power'],
                         driver_base.BareDriver().core_interfaces)
        self.assertEqual(
            ['bios', 'console', 'inspect', 'raid', 'rescue', 'storage'],
            driver_base.BareDriver().optional_interfaces
        )



# -*- encoding: utf-8 -*-
#
# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
SQLAlchemy models for baremetal data.
"""

from os import path
from urllib import parse as urlparse

from oslo_db import options as db_options
from oslo_db.sqlalchemy import models
from oslo_db.sqlalchemy import types as db_types
from sqlalchemy import Boolean, Column, DateTime, false, Index
from sqlalchemy import ForeignKey, Integer
from sqlalchemy import schema, String, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import orm

from ironic.common import exception
from ironic.common.i18n import _
from ironic.conf import CONF

_DEFAULT_SQL_CONNECTION = 'sqlite:///' + path.join('$state_path',
                                                   'ironic.sqlite')


db_options.set_defaults(CONF, connection=_DEFAULT_SQL_CONNECTION)


def table_args():
    engine_name = urlparse.urlparse(CONF.database.connection).scheme
    if engine_name == 'mysql':
        return {'mysql_engine': CONF.database.mysql_engine,
                'mysql_charset': "utf8"}
    return None


class IronicBase(models.TimestampMixin,
                 models.ModelBase):

    metadata = None

    version = Column(String(15), nullable=True)

    def as_dict(self):
        d = {}
        for c in self.__table__.columns:
            d[c.name] = self[c.name]
        return d


Base = declarative_base(cls=IronicBase)


class Chassis(Base):
    """Represents a hardware chassis."""

    __tablename__ = 'chassis'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_chassis0uuid'),
        table_args()
    )
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36))
    extra = Column(db_types.JsonEncodedDict)
    description = Column(String(255), nullable=True)


class Conductor(Base):
    """Represents a conductor service entry."""

    __tablename__ = 'conductors'
    __table_args__ = (
        schema.UniqueConstraint('hostname', name='uniq_conductors0hostname'),
        table_args()
    )
    id = Column(Integer, primary_key=True)
    hostname = Column(String(255), nullable=False)
    drivers = Column(db_types.JsonEncodedList)
    online = Column(Boolean, default=True)
    conductor_group = Column(String(255), nullable=False, default='',
                             server_default='')


class ConductorHardwareInterfaces(Base):
    """Internal table used to track what is loaded on each conductor."""

    __tablename__ = 'conductor_hardware_interfaces'
    __table_args__ = (
        schema.UniqueConstraint(
            'conductor_id',
            'hardware_type',
            'interface_type',
            'interface_name',
            name='uniq_conductorhardwareinterfaces0'),
        table_args())
    id = Column(Integer, primary_key=True)
    conductor_id = Column(Integer, ForeignKey('conductors.id'), nullable=False)
    hardware_type = Column(String(255), nullable=False)
    interface_type = Column(String(16), nullable=False)
    interface_name = Column(String(255), nullable=False)
    default = Column(Boolean, default=False, nullable=False)


class Node(Base):
    """Represents a bare metal node."""

    __tablename__ = 'nodes'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_nodes0uuid'),
        schema.UniqueConstraint('instance_uuid',
                                name='uniq_nodes0instance_uuid'),
        schema.UniqueConstraint('name', name='uniq_nodes0name'),
        Index('owner_idx', 'owner'),
        Index('lessee_idx', 'lessee'),
        Index('driver_idx', 'driver'),
        Index('provision_state_idx', 'provision_state'),
        Index('reservation_idx', 'reservation'),
        Index('conductor_group_idx', 'conductor_group'),
        Index('resource_class_idx', 'resource_class'),
        table_args())
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36))
    # NOTE(tenbrae): we store instance_uuid directly on the node so that we can
    #             filter on it more efficiently, even though it is
    #             user-settable, and would otherwise be in node.properties.
    instance_uuid = Column(String(36), nullable=True)
    name = Column(String(255), nullable=True)
    chassis_id = Column(Integer, ForeignKey('chassis.id'), nullable=True)
    power_state = Column(String(15), nullable=True)
    target_power_state = Column(String(15), nullable=True)
    provision_state = Column(String(15), nullable=True)
    target_provision_state = Column(String(15), nullable=True)
    provision_updated_at = Column(DateTime, nullable=True)
    last_error = Column(Text, nullable=True)
    instance_info = Column(db_types.JsonEncodedDict(mysql_as_long=True))
    properties = Column(db_types.JsonEncodedDict)
    driver = Column(String(255))
    driver_info = Column(db_types.JsonEncodedDict)
    driver_internal_info = Column(db_types.JsonEncodedDict)
    clean_step = Column(db_types.JsonEncodedDict)
    deploy_step = Column(db_types.JsonEncodedDict)
    resource_class = Column(String(80), nullable=True)

    raid_config = Column(db_types.JsonEncodedDict)
    target_raid_config = Column(db_types.JsonEncodedDict)

    # NOTE(tenbrae): this is the host name of the conductor which has
    #             acquired a TaskManager lock on the node.
    #             We should use an INT FK (conductors.id) in the future.
    reservation = Column(String(255), nullable=True)

    # NOTE(tenbrae): this is the id of the last conductor which prepared local
    #             state for the node (eg, a PXE config file).
    #             When affinity and the hash ring's mapping do not match,
    #             this indicates that a conductor should rebuild local state.
    conductor_affinity = Column(Integer,
                                ForeignKey('conductors.id',
                                           name='nodes_conductor_affinity_fk'),
                                nullable=True)
    conductor_group = Column(String(255), nullable=False, default='',
                             server_default='')

    maintenance = Column(Boolean, default=False)
    maintenance_reason = Column(Text, nullable=True)
    fault = Column(String(255), nullable=True)
    console_enabled = Column(Boolean, default=False)
    inspection_finished_at = Column(DateTime, nullable=True)
    inspection_started_at = Column(DateTime, nullable=True)
    extra = Column(db_types.JsonEncodedDict)
    automated_clean = Column(Boolean, nullable=True)
    protected = Column(Boolean, nullable=False, default=False,
                       server_default=false())
    protected_reason = Column(Text, nullable=True)
    owner = Column(String(255), nullable=True)
    lessee = Column(String(255), nullable=True)
    allocation_id = Column(Integer, ForeignKey('allocations.id'),
                           nullable=True)
    description = Column(Text, nullable=True)

    bios_interface = Column(String(255), nullable=True)
    boot_interface = Column(String(255), nullable=True)
    console_interface = Column(String(255), nullable=True)
    deploy_interface = Column(String(255), nullable=True)
    inspect_interface = Column(String(255), nullable=True)
    management_interface = Column(String(255), nullable=True)
    network_interface = Column(String(255), nullable=True)
    raid_interface = Column(String(255), nullable=True)
    rescue_interface = Column(String(255), nullable=True)
    retired = Column(Boolean, nullable=True, default=False,
                     server_default=false())
    retired_reason = Column(Text, nullable=True)
    network_data = Column(db_types.JsonEncodedDict)
    storage_interface = Column(String(255), nullable=True)
    power_interface = Column(String(255), nullable=True)
    vendor_interface = Column(String(255), nullable=True)

    boot_mode = Column(String(16), nullable=True)
    secure_boot = Column(Boolean, nullable=True)


class Port(Base):
    """Represents a network port of a bare metal node."""

    __tablename__ = 'ports'
    __table_args__ = (
        schema.UniqueConstraint('address', name='uniq_ports0address'),
        schema.UniqueConstraint('uuid', name='uniq_ports0uuid'),
        schema.UniqueConstraint('name', name='uniq_ports0name'),
        table_args())
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36))
    address = Column(String(18))
    node_id = Column(Integer, ForeignKey('nodes.id'), nullable=True)
    extra = Column(db_types.JsonEncodedDict)
    local_link_connection = Column(db_types.JsonEncodedDict)
    portgroup_id = Column(Integer, ForeignKey('portgroups.id'), nullable=True)
    pxe_enabled = Column(Boolean, default=True)
    internal_info = Column(db_types.JsonEncodedDict)
    physical_network = Column(String(64), nullable=True)
    is_smartnic = Column(Boolean, nullable=True, default=False)
    name = Column(String(255), nullable=True)


class Portgroup(Base):
    """Represents a group of network ports of a bare metal node."""

    __tablename__ = 'portgroups'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_portgroups0uuid'),
        schema.UniqueConstraint('address', name='uniq_portgroups0address'),
        schema.UniqueConstraint('name', name='uniq_portgroups0name'),
        table_args())
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36))
    name = Column(String(255), nullable=True)
    node_id = Column(Integer, ForeignKey('nodes.id'), nullable=True)
    address = Column(String(18))
    extra = Column(db_types.JsonEncodedDict)
    internal_info = Column(db_types.JsonEncodedDict)
    standalone_ports_supported = Column(Boolean, default=True)
    mode = Column(String(255))
    properties = Column(db_types.JsonEncodedDict)


class NodeTag(Base):
    """Represents a tag of a bare metal node."""

    __tablename__ = 'node_tags'
    __table_args__ = (
        Index('node_tags_idx', 'tag'),
        table_args())
    node_id = Column(Integer, ForeignKey('nodes.id'),
                     primary_key=True, nullable=False)
    tag = Column(String(255), primary_key=True, nullable=False)

    node = orm.relationship(
        "Node",
        backref='tags',
        primaryjoin='and_(NodeTag.node_id == Node.id)',
        foreign_keys=node_id
    )


class VolumeConnector(Base):
    """Represents a volume connector of a bare metal node."""

    __tablename__ = 'volume_connectors'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_volumeconnectors0uuid'),
        schema.UniqueConstraint(
            'type',
            'connector_id',
            name='uniq_volumeconnectors0type0connector_id'),
        table_args())
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36))
    node_id = Column(Integer, ForeignKey('nodes.id'), nullable=True)
    type = Column(String(32))
    connector_id = Column(String(255))
    extra = Column(db_types.JsonEncodedDict)


class VolumeTarget(Base):
    """Represents a volume target of a bare metal node."""

    __tablename__ = 'volume_targets'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_volumetargets0uuid'),
        schema.UniqueConstraint('node_id',
                                'boot_index',
                                name='uniq_volumetargets0node_id0boot_index'),
        table_args())
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36))
    node_id = Column(Integer, ForeignKey('nodes.id'), nullable=True)
    volume_type = Column(String(64))
    properties = Column(db_types.JsonEncodedDict)
    boot_index = Column(Integer)
    volume_id = Column(String(36))
    extra = Column(db_types.JsonEncodedDict)


class NodeTrait(Base):
    """Represents a trait of a bare metal node."""

    __tablename__ = 'node_traits'
    __table_args__ = (
        Index('node_traits_idx', 'trait'),
        table_args())
    node_id = Column(Integer, ForeignKey('nodes.id'),
                     primary_key=True, nullable=False)
    trait = Column(String(255), primary_key=True, nullable=False)
    node = orm.relationship(
        "Node",
        backref='traits',
        primaryjoin='and_(NodeTrait.node_id == Node.id)',
        foreign_keys=node_id
    )


class BIOSSetting(Base):
    """Represents a bios setting of a bare metal node."""

    __tablename__ = 'bios_settings'
    __table_args__ = (table_args())
    node_id = Column(Integer, ForeignKey('nodes.id'),
                     primary_key=True, nullable=False)
    name = Column(String(255), primary_key=True, nullable=False)
    value = Column(Text, nullable=True)
    attribute_type = Column(String(255), nullable=True)
    allowable_values = Column(db_types.JsonEncodedList, nullable=True)
    lower_bound = Column(Integer, nullable=True)
    max_length = Column(Integer, nullable=True)
    min_length = Column(Integer, nullable=True)
    read_only = Column(Boolean, nullable=True)
    reset_required = Column(Boolean, nullable=True)
    unique = Column(Boolean, nullable=True)
    upper_bound = Column(Integer, nullable=True)


class Allocation(Base):
    """Represents an allocation of a node for deployment."""

    __tablename__ = 'allocations'
    __table_args__ = (
        schema.UniqueConstraint('name', name='uniq_allocations0name'),
        schema.UniqueConstraint('uuid', name='uniq_allocations0uuid'),
        table_args())
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36), nullable=False)
    name = Column(String(255), nullable=True)
    node_id = Column(Integer, ForeignKey('nodes.id'), nullable=True)
    state = Column(String(15), nullable=False)
    owner = Column(String(255), nullable=True)
    last_error = Column(Text, nullable=True)
    resource_class = Column(String(80), nullable=True)
    traits = Column(db_types.JsonEncodedList)
    candidate_nodes = Column(db_types.JsonEncodedList)
    extra = Column(db_types.JsonEncodedDict)
    # The last conductor to handle this allocation (internal field).
    conductor_affinity = Column(Integer, ForeignKey('conductors.id'),
                                nullable=True)


class DeployTemplate(Base):
    """Represents a deployment template."""

    __tablename__ = 'deploy_templates'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_deploytemplates0uuid'),
        schema.UniqueConstraint('name', name='uniq_deploytemplates0name'),
        table_args())
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36))
    name = Column(String(255), nullable=False)
    extra = Column(db_types.JsonEncodedDict)


class DeployTemplateStep(Base):
    """Represents a deployment step in a deployment template."""

    __tablename__ = 'deploy_template_steps'
    __table_args__ = (
        Index('deploy_template_id', 'deploy_template_id'),
        Index('deploy_template_steps_interface_idx', 'interface'),
        Index('deploy_template_steps_step_idx', 'step'),
        table_args())
    id = Column(Integer, primary_key=True)
    deploy_template_id = Column(Integer, ForeignKey('deploy_templates.id'),
                                nullable=False)
    interface = Column(String(255), nullable=False)
    step = Column(String(255), nullable=False)
    args = Column(db_types.JsonEncodedDict, nullable=False)
    priority = Column(Integer, nullable=False)
    deploy_template = orm.relationship(
        "DeployTemplate",
        backref='steps',
        primaryjoin=(
            'and_(DeployTemplateStep.deploy_template_id == '
            'DeployTemplate.id)'),
        foreign_keys=deploy_template_id
    )


class NodeHistory(Base):
    """Represents a history event of a bare metal node."""

    __tablename__ = 'node_history'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_history0uuid'),
        Index('history_node_id_idx', 'node_id'),
        Index('history_uuid_idx', 'uuid'),
        Index('history_conductor_idx', 'conductor'),
        table_args())
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36), nullable=False)
    conductor = Column(String(255), nullable=True)
    event_type = Column(String(255), nullable=True)
    severity = Column(String(255), nullable=True)
    event = Column(Text, nullable=True)
    user = Column(String(32), nullable=True)
    node_id = Column(Integer, ForeignKey('nodes.id'), nullable=True)


def get_class(model_name):
    """Returns the model class with the specified name.

    :param model_name: the name of the class
    :returns: the class with the specified name
    :raises: Exception if there is no class associated with the name
    """
    for model in Base.__subclasses__():
        if model.__name__ == model_name:
            return model

    raise exception.IronicException(
        _("Cannot find model with name: %s") % model_name)



# Copyright 2015 Rackspace, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg
from oslo_log import log

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import neutron
from ironic.common import states
from ironic.drivers import base
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.network import common

LOG = log.getLogger(__name__)

CONF = cfg.CONF


class NeutronNetwork(common.NeutronVIFPortIDMixin,
                     neutron.NeutronNetworkInterfaceMixin,
                     base.NetworkInterface):
    """Neutron v2 network interface"""

    def __init__(self):
        failures = []
        cleaning_net = CONF.neutron.cleaning_network
        if not cleaning_net:
            failures.append('cleaning_network')

        provisioning_net = CONF.neutron.provisioning_network
        if not provisioning_net:
            failures.append('provisioning_network')

        if failures:
            raise exception.DriverLoadError(
                driver=self.__class__.__name__,
                reason=(_('The following [neutron] group configuration '
                          'options are missing: %s') % ', '.join(failures)))

    def validate(self, task):
        """Validates the network interface.

        :param task: a TaskManager instance.
        :raises: InvalidParameterValue, if the network interface configuration
            is invalid.
        :raises: MissingParameterValue, if some parameters are missing.
        """
        self.get_cleaning_network_uuid(task)
        self.get_provisioning_network_uuid(task)
        node = task.node
        if (node.provision_state == states.DEPLOYING
            and node.driver_internal_info.get('is_whole_disk_image')
            and deploy_utils.get_boot_option(node) == 'netboot'):
            error_msg = (_('The node %s cannot perform "local" boot for '
                           'whole disk image when node is using "neutron" '
                           'network and is configured with "netboot" boot '
                           'option.') % node.uuid)
            raise exception.InvalidParameterValue(error_msg)

    def _add_network(self, task, network, security_groups, process):
        # If we have left over ports from a previous process, remove them
        neutron.rollback_ports(task, network)
        LOG.info('Adding %s network to node %s', process, task.node.uuid)
        vifs = neutron.add_ports_to_network(task, network,
                                            security_groups=security_groups)
        field = '%s_vif_port_id' % process
        for port in task.ports:
            if port.uuid in vifs:
                internal_info = port.internal_info
                internal_info[field] = vifs[port.uuid]
                port.internal_info = internal_info
                port.save()
        return vifs

    def _remove_network(self, task, network, process):
        LOG.info('Removing ports from %s network for node %s',
                 process, task.node.uuid)
        neutron.remove_ports_from_network(task, network)
        field = '%s_vif_port_id' % process
        for port in task.ports:
            if field in port.internal_info:
                internal_info = port.internal_info
                del internal_info[field]
                port.internal_info = internal_info
                port.save()

    def add_provisioning_network(self, task):
        """Add the provisioning network to a node.

        :param task: A TaskManager instance.
        :raises: NetworkError
        """
        self._add_network(
            task, self.get_provisioning_network_uuid(task),
            CONF.neutron.provisioning_network_security_groups,
            'provisioning')

    def remove_provisioning_network(self, task):
        """Remove the provisioning network from a node.

        :param task: A TaskManager instance.
        :raises: NetworkError
        """
        return self._remove_network(
            task, self.get_provisioning_network_uuid(task), 'provisioning')

    def add_cleaning_network(self, task):
        """Create neutron ports for each port on task.node to boot the ramdisk.

        :param task: a TaskManager instance.
        :raises: NetworkError
        :returns: a dictionary in the form {port.uuid: neutron_port['id']}
        """
        return self._add_network(
            task, self.get_cleaning_network_uuid(task),
            CONF.neutron.cleaning_network_security_groups,
            'cleaning')

    def remove_cleaning_network(self, task):
        """Deletes the neutron port created for booting the ramdisk.

        :param task: a TaskManager instance.
        :raises: NetworkError
        """
        return self._remove_network(
            task, self.get_cleaning_network_uuid(task), 'cleaning')

    def validate_rescue(self, task):
        """Validates the network interface for rescue operation.

        :param task: a TaskManager instance.
        :raises: InvalidParameterValue, if the network interface configuration
            is invalid.
        :raises: MissingParameterValue, if some parameters are missing.
        """
        self.get_rescuing_network_uuid(task)

    def add_rescuing_network(self, task):
        """Create neutron ports for each port to boot the rescue ramdisk.

        :param task: a TaskManager instance.
        :returns: a dictionary in the form {port.uuid: neutron_port['id']}
        """
        return self._add_network(
            task, self.get_rescuing_network_uuid(task),
            CONF.neutron.rescuing_network_security_groups,
            'rescuing')

    def remove_rescuing_network(self, task):
        """Deletes neutron port created for booting the rescue ramdisk.

        :param task: a TaskManager instance.
        :raises: NetworkError
        """
        return self._remove_network(
            task, self.get_rescuing_network_uuid(task), 'rescuing')

    def configure_tenant_networks(self, task):
        """Configure tenant networks for a node.

        :param task: A TaskManager instance.
        :raises: NetworkError
        """
        node = task.node
        ports = task.ports
        LOG.info('Mapping instance ports to %s', node.uuid)

        # TODO(russell_h): this is based on the broken assumption that the
        # number of Neutron ports will match the number of physical ports.
        # Instead, we should probably list ports for this instance in
        # Neutron and update all of those with the appropriate portmap.
        if not ports:
            msg = _("No ports are associated with node %s") % node.uuid
            LOG.error(msg)
            raise exception.NetworkError(msg)
        ports = [p for p in ports if not p.portgroup_id]
        portgroups = task.portgroups

        client = neutron.get_client(context=task.context)
        pobj_without_vif = 0
        for port_like_obj in ports + portgroups:

            try:
                common.plug_port_to_tenant_network(task, port_like_obj,
                                                   client=client)
            except exception.VifNotAttached:
                pobj_without_vif += 1
                continue

        if pobj_without_vif == len(ports + portgroups):
            msg = _("No neutron ports or portgroups are associated with "
                    "node %s") % node.uuid
            LOG.error(msg)
            raise exception.NetworkError(msg)

    def unconfigure_tenant_networks(self, task):
        """Unconfigure tenant networks for a node.

        Nova takes care of port removal from tenant network, we unbind it
        here/now to avoid the possibility of the ironic port being bound to the
        tenant and cleaning networks at the same time.

        :param task: A TaskManager instance.
        :raises: NetworkError
        """
        node = task.node
        LOG.info('Unbinding instance ports from node %s', node.uuid)

        ports = [p for p in task.ports if not p.portgroup_id]
        portgroups = task.portgroups
        for port_like_obj in ports + portgroups:
            vif_port_id = (
                port_like_obj.internal_info.get(common.TENANT_VIF_KEY)
            )
            if not vif_port_id:
                continue

            is_smart_nic = neutron.is_smartnic_port(port_like_obj)
            if is_smart_nic:
                client = neutron.get_client(context=task.context)
                link_info = port_like_obj.local_link_connection
                neutron.wait_for_host_agent(client, link_info['hostname'])

            # NOTE(kaifeng) address is optional for port group, avoid to
            # regenerate mac when the address is absent.
            reset_mac = bool(port_like_obj.address)
            neutron.unbind_neutron_port(vif_port_id, context=task.context,
                                        reset_mac=reset_mac)

    def need_power_on(self, task):
        """Check if the node has any Smart NIC ports

        :param task: A TaskManager instance.
        :return: A boolean to indicate Smart NIC port presence
        """
        for port in task.ports:
            if neutron.is_smartnic_port(port):
                return True
        return False

    def add_inspection_network(self, task):
        """Add the inspection network to the node.

        :param task: A TaskManager instance.
        :returns: a dictionary in the form {port.uuid: neutron_port['id']}
        :raises: NetworkError
        :raises: InvalidParameterValue, if the network interface configuration
            is invalid.
        """
        return self._add_network(
            task, self.get_inspection_network_uuid(task),
            CONF.neutron.inspection_network_security_groups,
            'inspection')

    def remove_inspection_network(self, task):
        """Removes the inspection network from a node.

        :param task: A TaskManager instance.
        :raises: InvalidParameterValue, if the network interface configuration
            is invalid.
        :raises: MissingParameterValue, if some parameters are missing.
        """
        return self._remove_network(
            task, self.get_inspection_network_uuid(task), 'inspection')



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
DRAC Driver for remote system management using Dell Remote Access Card.
"""

from oslo_config import cfg

from ironic.drivers import generic
from ironic.drivers.modules.drac import bios
from ironic.drivers.modules.drac import boot
from ironic.drivers.modules.drac import inspect as drac_inspect
from ironic.drivers.modules.drac import management
from ironic.drivers.modules.drac import power
from ironic.drivers.modules.drac import raid
from ironic.drivers.modules.drac import vendor_passthru
from ironic.drivers.modules import ipxe
from ironic.drivers.modules import noop
from ironic.drivers.modules import pxe


CONF = cfg.CONF


class IDRACHardware(generic.GenericHardware):
    """integrated Dell Remote Access Controller hardware type"""

    # Required hardware interfaces

    @property
    def supported_boot_interfaces(self):
        """List of supported boot interfaces."""
        return [ipxe.iPXEBoot, pxe.PXEBoot, boot.DracRedfishVirtualMediaBoot]

    @property
    def supported_management_interfaces(self):
        """List of supported management interfaces."""
        return [management.DracWSManManagement, management.DracManagement,
                management.DracRedfishManagement]

    @property
    def supported_power_interfaces(self):
        """List of supported power interfaces."""
        return [power.DracWSManPower, power.DracPower, power.DracRedfishPower]

    # Optional hardware interfaces

    @property
    def supported_bios_interfaces(self):
        """List of supported bios interfaces."""
        return [bios.DracWSManBIOS, bios.DracRedfishBIOS, noop.NoBIOS]

    @property
    def supported_inspect_interfaces(self):
        """List of supported inspect interfaces."""
        # Inspector support should have a higher priority than NoInspect
        # if it is enabled by an operator (implying that the service is
        # installed).
        return [drac_inspect.DracWSManInspect, drac_inspect.DracInspect,
                drac_inspect.DracRedfishInspect] + super(
                    IDRACHardware, self).supported_inspect_interfaces

    @property
    def supported_raid_interfaces(self):
        """List of supported raid interfaces."""
        return [raid.DracWSManRAID, raid.DracRAID,
                raid.DracRedfishRAID] + super(
                    IDRACHardware, self).supported_raid_interfaces

    @property
    def supported_vendor_interfaces(self):
        """List of supported vendor interfaces."""
        return [vendor_passthru.DracWSManVendorPassthru,
                vendor_passthru.DracVendorPassthru,
                vendor_passthru.DracRedfishVendorPassthru, noop.NoVendor]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating portgroups via the DB API"""

from oslo_utils import uuidutils

from ironic.common import exception
from ironic.tests.unit.db import base
from ironic.tests.unit.db import utils as db_utils


class DbportgroupTestCase(base.DbTestCase):

    def setUp(self):
        # This method creates a portgroup for every test and
        # replaces a test for creating a portgroup.
        super(DbportgroupTestCase, self).setUp()
        self.node = db_utils.create_test_node()
        self.portgroup = db_utils.create_test_portgroup(node_id=self.node.id)

    def _create_test_portgroup_range(self, count):
        """Create the specified number of test portgroup entries in DB

        It uses create_test_portgroup method. And returns List of Portgroup
        DB objects.

        :param count: Specifies the number of portgroups to be created
        :returns: List of Portgroup DB objects

        """
        uuids = []
        for i in range(1, count):
            portgroup = db_utils.create_test_portgroup(
                uuid=uuidutils.generate_uuid(),
                node_id=self.node.id,
                name='portgroup' + str(i),
                address='52:54:00:cf:2d:4%s' % i)
            uuids.append(str(portgroup.uuid))

        return uuids

    def test_get_portgroup_by_id(self):
        res = self.dbapi.get_portgroup_by_id(self.portgroup.id)
        self.assertEqual(self.portgroup.address, res.address)

    def test_get_portgroup_by_id_that_does_not_exist(self):
        self.assertRaises(exception.PortgroupNotFound,
                          self.dbapi.get_portgroup_by_id, 99)

    def test_get_portgroup_by_uuid(self):
        res = self.dbapi.get_portgroup_by_uuid(self.portgroup.uuid)
        self.assertEqual(self.portgroup.id, res.id)

    def test_get_portgroup_by_uuid_that_does_not_exist(self):
        self.assertRaises(exception.PortgroupNotFound,
                          self.dbapi.get_portgroup_by_uuid,
                          'EEEEEEEE-EEEE-EEEE-EEEE-EEEEEEEEEEEE')

    def test_get_portgroup_by_address(self):
        res = self.dbapi.get_portgroup_by_address(self.portgroup.address)
        self.assertEqual(self.portgroup.id, res.id)

    def test_get_portgroup_by_address_that_does_not_exist(self):
        self.assertRaises(exception.PortgroupNotFound,
                          self.dbapi.get_portgroup_by_address,
                          '31:31:31:31:31:31')

    def test_get_portgroup_by_name(self):
        res = self.dbapi.get_portgroup_by_name(self.portgroup.name)
        self.assertEqual(self.portgroup.id, res.id)

    def test_get_portgroup_by_name_that_does_not_exist(self):
        self.assertRaises(exception.PortgroupNotFound,
                          self.dbapi.get_portgroup_by_name, 'testfail')

    def test_get_portgroup_list(self):
        uuids = self._create_test_portgroup_range(6)

        # Also add the uuid for the portgroup created in setUp()
        uuids.append(str(self.portgroup.uuid))
        res = self.dbapi.get_portgroup_list()
        res_uuids = [r.uuid for r in res]
        self.assertCountEqual(uuids, res_uuids)

    def test_get_portgroup_list_sorted(self):
        uuids = self._create_test_portgroup_range(6)

        # Also add the uuid for the portgroup created in setUp()
        uuids.append(str(self.portgroup.uuid))
        res = self.dbapi.get_portgroup_list(sort_key='uuid')
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), res_uuids)

        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.get_portgroup_list, sort_key='foo')

    def test_get_portgroups_by_node_id(self):
        res = self.dbapi.get_portgroups_by_node_id(self.node.id)
        self.assertEqual(self.portgroup.address, res[0].address)

    def test_get_portgroups_by_node_id_that_does_not_exist(self):
        self.assertEqual([], self.dbapi.get_portgroups_by_node_id(99))

    def test_destroy_portgroup(self):
        self.dbapi.destroy_portgroup(self.portgroup.id)
        self.assertRaises(exception.PortgroupNotFound,
                          self.dbapi.get_portgroup_by_id, self.portgroup.id)

    def test_destroy_portgroup_that_does_not_exist(self):
        self.assertRaises(exception.PortgroupNotFound,
                          self.dbapi.destroy_portgroup, 99)

    def test_destroy_portgroup_uuid(self):
        self.dbapi.destroy_portgroup(self.portgroup.uuid)

    def test_destroy_portgroup_not_empty(self):
        self.port = db_utils.create_test_port(node_id=self.node.id,
                                              portgroup_id=self.portgroup.id)
        self.assertRaises(exception.PortgroupNotEmpty,
                          self.dbapi.destroy_portgroup, self.portgroup.id)

    def test_update_portgroup(self):
        old_address = self.portgroup.address
        new_address = 'ff:ee:dd:cc:bb:aa'
        self.assertNotEqual(old_address, new_address)
        old_name = self.portgroup.name
        new_name = 'newname'
        self.assertNotEqual(old_name, new_name)
        res = self.dbapi.update_portgroup(self.portgroup.id,
                                          {'address': new_address,
                                           'name': new_name})
        self.assertEqual(new_address, res.address)
        self.assertEqual(new_name, res.name)

    def test_update_portgroup_uuid(self):
        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.update_portgroup, self.portgroup.id,
                          {'uuid': ''})

    def test_update_portgroup_not_found(self):
        id_2 = 99
        self.assertNotEqual(self.portgroup.id, id_2)
        address2 = 'aa:bb:cc:11:22:33'
        self.assertRaises(exception.PortgroupNotFound,
                          self.dbapi.update_portgroup, id_2,
                          {'address': address2})

    def test_update_portgroup_duplicated_address(self):
        address1 = self.portgroup.address
        address2 = 'aa:bb:cc:11:22:33'
        portgroup2 = db_utils.create_test_portgroup(
            uuid=uuidutils.generate_uuid(),
            node_id=self.node.id,
            name=str(uuidutils.generate_uuid()),
            address=address2)
        self.assertRaises(exception.PortgroupMACAlreadyExists,
                          self.dbapi.update_portgroup, portgroup2.id,
                          {'address': address1})

    def test_update_portgroup_duplicated_name(self):
        name1 = self.portgroup.name
        portgroup2 = db_utils.create_test_portgroup(
            uuid=uuidutils.generate_uuid(),
            node_id=self.node.id,
            name='name2', address='aa:bb:cc:11:22:55')
        self.assertRaises(exception.PortgroupDuplicateName,
                          self.dbapi.update_portgroup, portgroup2.id,
                          {'name': name1})

    def test_create_portgroup_duplicated_name(self):
        self.assertRaises(exception.PortgroupDuplicateName,
                          db_utils.create_test_portgroup,
                          uuid=uuidutils.generate_uuid(),
                          node_id=self.node.id,
                          name=self.portgroup.name,
                          address='aa:bb:cc:11:22:55')

    def test_create_portgroup_duplicated_address(self):
        self.assertRaises(exception.PortgroupMACAlreadyExists,
                          db_utils.create_test_portgroup,
                          uuid=uuidutils.generate_uuid(),
                          node_id=self.node.id,
                          name=str(uuidutils.generate_uuid()),
                          address=self.portgroup.address)

    def test_create_portgroup_duplicated_uuid(self):
        self.assertRaises(exception.PortgroupAlreadyExists,
                          db_utils.create_test_portgroup,
                          uuid=self.portgroup.uuid,
                          node_id=self.node.id,
                          name=str(uuidutils.generate_uuid()),
                          address='aa:bb:cc:33:11:22')

    def test_create_portgroup_no_mode(self):
        self.config(default_portgroup_mode='802.3ad')
        name = uuidutils.generate_uuid()
        db_utils.create_test_portgroup(uuid=uuidutils.generate_uuid(),
                                       node_id=self.node.id, name=name,
                                       address='aa:bb:cc:dd:ee:ff')
        res = self.dbapi.get_portgroup_by_id(self.portgroup.id)
        self.assertEqual('active-backup', res.mode)
        res = self.dbapi.get_portgroup_by_name(name)
        self.assertEqual('802.3ad', res.mode)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Run storage database migration.
"""

import sys

from oslo_config import cfg

from ironic.common import context
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import service
from ironic.conf import CONF
from ironic.db import api as db_api
from ironic.db import migration
from ironic import version


dbapi = db_api.get_instance()

# NOTE(rloo): This is a list of functions to perform online data migrations
# (from previous releases) for this release, in batches. It may be empty.
# The migration functions should be ordered by execution order; from earlier
# to later releases.
#
# Each migration function takes two arguments -- the context and maximum
# number of objects to migrate, and returns a 2-tuple -- the total number of
# objects that need to be migrated at the beginning of the function, and the
# number migrated. If the function determines that no migrations are needed,
# it returns (0, 0).
#
# The last migration step should always remain the last one -- it migrates
# all objects to their latest known versions.
#
# Example of a function docstring:
#
#        def sample_data_migration(context, max_count):
#        """Sample method to migrate data to new format.
#
#        :param context: an admin context
#        :param max_count: The maximum number of objects to migrate. Must be
#                          >= 0. If zero, all the objects will be migrated.
#        :returns: A 2-tuple -- the total number of objects that need to be
#                  migrated (at the beginning of this call) and the number
#                  of migrated objects.
#        """
# NOTE(vdrok): Do not access objects' attributes, instead only provide object
# and attribute name tuples, so that not to trigger the load of the whole
# object, in case it is lazy loaded. The attribute will be accessed when needed
# by doing getattr on the object
ONLINE_MIGRATIONS = (
    # NOTE(rloo): Don't remove this; it should always be last
    (dbapi, 'update_to_latest_versions'),
)


class DBCommand(object):

    def check_obj_versions(self, ignore_missing_tables=False):
        """Check the versions of objects.

        Check that the object versions are compatible with this release
        of ironic. It does this by comparing the objects' .version field
        in the database, with the expected versions of these objects.

        Returns None if compatible; a string describing the issue otherwise.
        """
        if migration.version() is None:
            # no tables, nothing to check
            return

        msg = None
        try:
            if not dbapi.check_versions(
                    permit_initial_version=ignore_missing_tables):
                msg = (_('The database is not compatible with this '
                         'release of ironic (%s). Please run '
                         '"ironic-dbsync online_data_migrations" using '
                         'the previous release.\n')
                       % version.version_info.release_string())
        except exception.DatabaseVersionTooOld:
            msg = (_('The database version is not compatible with this '
                     'release of ironic (%s). This can happen if you are '
                     'attempting to upgrade from a version older than '
                     'the previous release (skip versions upgrade). '
                     'This is an unsupported upgrade method. '
                     'Please run "ironic-dbsync upgrade" using the previous '
                     'releases for a fast-forward upgrade.\n')
                   % version.version_info.release_string())

        return msg

    def _check_versions(self, ignore_missing_tables=False):
        msg = self.check_obj_versions(
            ignore_missing_tables=ignore_missing_tables)
        if not msg:
            return
        else:
            sys.stderr.write(msg)
            # NOTE(rloo): We return 1 in online_data_migrations() to indicate
            # that there are more objects to migrate, so don't use 1 here.
            sys.exit(2)

    def upgrade(self):
        self._check_versions(ignore_missing_tables=True)
        migration.upgrade(CONF.command.revision)

    def revision(self):
        migration.revision(CONF.command.message, CONF.command.autogenerate)

    def stamp(self):
        migration.stamp(CONF.command.revision)

    def version(self):
        print(migration.version())

    def create_schema(self):
        migration.create_schema()

    def online_data_migrations(self):
        self._check_versions()
        self._run_online_data_migrations(max_count=CONF.command.max_count,
                                         options=CONF.command.options)

    def _run_migration_functions(self, context, max_count, options):
        """Runs the migration functions.

        Runs the data migration functions in the ONLINE_MIGRATIONS list.
        It makes sure the total number of object migrations doesn't exceed the
        specified max_count. A migration of an object will typically migrate
        one row of data inside the database.

        :param context: an admin context
        :param max_count: the maximum number of objects (rows) to migrate;
            a value >= 1.
        :param options: migration options - dict mapping migration name
            to a dictionary of options for this migration.
        :raises: Exception from the migration function
        :returns: Boolean value indicating whether migrations are done. Returns
            False if max_count objects have been migrated (since at that
            point, it is unknown whether all migrations are done). Returns
            True if migrations are all done (i.e. fewer than max_count objects
            were migrated when the migrations are done).
        """
        total_migrated = 0

        for migration_func_obj, migration_func_name in ONLINE_MIGRATIONS:
            migration_func = getattr(migration_func_obj, migration_func_name)
            migration_opts = options.get(migration_func_name, {})
            num_to_migrate = max_count - total_migrated
            try:
                total_to_do, num_migrated = migration_func(context,
                                                           num_to_migrate,
                                                           **migration_opts)
            except Exception as e:
                print(_("Error while running %(migration)s: %(err)s.")
                      % {'migration': migration_func.__name__, 'err': e},
                      file=sys.stderr)
                raise

            print(_('%(migration)s() migrated %(done)i of %(total)i objects.')
                  % {'migration': migration_func.__name__,
                     'total': total_to_do,
                     'done': num_migrated})
            total_migrated += num_migrated
            if total_migrated >= max_count:
                # NOTE(rloo). max_count objects have been migrated so we have
                # to stop. We return False because there is no look-ahead so
                # we don't know if the migrations have been all done. All we
                # know is that we've migrated max_count. It is possible that
                # the migrations are done and that there aren't any more to
                # migrate after this, but that would involve checking:
                #   1. num_migrated == total_to_do (easy enough), AND
                #   2. whether there are other migration functions and whether
                #      they need to do any object migrations (not so easy to
                #      check)
                return False

        return True

    def _run_online_data_migrations(self, max_count=None, options=None):
        """Perform online data migrations for the release.

        Online data migrations are done by running all the data migration
        functions in the ONLINE_MIGRATIONS list. If max_count is None, all
        the functions will be run in batches of 50 objects, until the
        migrations are done. Otherwise, this will run (some of) the functions
        until max_count objects have been migrated.

        :param max_count: the maximum number of individual object migrations
            or modified rows, a value >= 1. If None, migrations are run in a
            loop in batches of 50, until completion.
        :param options: options to pass to migrations. List of values in the
            form of <migration name>.<option>=<value>
        :raises: SystemExit. With exit code of:
            0: when all migrations are complete.
            1: when objects were migrated and the command needs to be
               re-run (because there might be more objects to be migrated)
            127: if max_count is < 1 or any option is invalid
        :raises: Exception from a migration function
        """
        parsed_options = {}
        if options:
            for option in options:
                try:
                    migration, key_value = option.split('.', 1)
                    key, value = key_value.split('=', 1)
                except ValueError:
                    print(_("Malformed option %s") % option)
                    sys.exit(127)
                else:
                    parsed_options.setdefault(migration, {})[key] = value

        admin_context = context.get_admin_context()
        finished_migrating = False
        if max_count is None:
            max_count = 50
            print(_('Running batches of %i until migrations have been '
                    'completed.') % max_count)
            while not finished_migrating:
                finished_migrating = self._run_migration_functions(
                    admin_context, max_count, parsed_options)
            print(_('Data migrations have completed.'))
            sys.exit(0)

        if max_count < 1:
            print(_('"max-count" must be a positive value.'), file=sys.stderr)
            sys.exit(127)

        finished_migrating = self._run_migration_functions(admin_context,
                                                           max_count,
                                                           parsed_options)
        if finished_migrating:
            print(_('Data migrations have completed.'))
            sys.exit(0)
        else:
            print(_('Data migrations have not completed. Please re-run.'))
            sys.exit(1)


def add_command_parsers(subparsers):
    command_object = DBCommand()

    parser = subparsers.add_parser(
        'upgrade',
        help=_("Upgrade the database schema to the latest version. "
               "Optionally, use --revision to specify an alembic revision "
               "string to upgrade to. It returns 2 (error) if the database is "
               "not compatible with this version. If this happens, the "
               "'ironic-dbsync online_data_migrations' command should be run "
               "using the previous version of ironic, before upgrading and "
               "running this command."))

    parser.set_defaults(func=command_object.upgrade)
    parser.add_argument('--revision', nargs='?')

    parser = subparsers.add_parser('stamp')
    parser.add_argument('--revision', nargs='?')
    parser.set_defaults(func=command_object.stamp)

    parser = subparsers.add_parser(
        'revision',
        help=_("Create a new alembic revision. "
               "Use --message to set the message string."))
    parser.add_argument('-m', '--message')
    parser.add_argument('--autogenerate', action='store_true')
    parser.set_defaults(func=command_object.revision)

    parser = subparsers.add_parser(
        'version',
        help=_("Print the current version information and exit."))
    parser.set_defaults(func=command_object.version)

    parser = subparsers.add_parser(
        'create_schema',
        help=_("Create the database schema."))
    parser.set_defaults(func=command_object.create_schema)

    parser = subparsers.add_parser(
        'online_data_migrations',
        help=_("Perform online data migrations for the release. If "
               "--max-count is specified, at most max-count objects will be "
               "migrated. If not specified, all objects will be migrated "
               "(in batches to avoid locking the database for long periods of "
               "time). "
               "The command returns code 0 (success) after migrations are "
               "finished or there are no data to migrate. It returns code "
               "1 (error) if there are still pending objects to be migrated. "
               "Before upgrading to a newer release, this command must be run "
               "until code 0 is returned. "
               "It returns 127 (error) if max-count is < 1. "
               "It returns 2 (error) if the database is not compatible with "
               "this release. If this happens, this command should be run "
               "using the previous release of ironic, before upgrading and "
               "running this command."))
    parser.add_argument(
        '--max-count', metavar='<number>', dest='max_count', type=int,
        help=_("Maximum number of objects to migrate. If unspecified, all "
               "objects are migrated."))
    parser.add_argument(
        '--option', metavar='<migration.opt=val>', action='append',
        dest='options', default=[],
        help=_("Options to pass to the migrations in the form of "
               "<migration name>.<option>=<value>"))
    parser.set_defaults(func=command_object.online_data_migrations)


def main():
    command_opt = cfg.SubCommandOpt('command',
                                    title='Command',
                                    help=_('Available commands'),
                                    handler=add_command_parsers)

    CONF.register_cli_opt(command_opt)

    # this is hack to work with previous usage of ironic-dbsync
    # pls change it to ironic-dbsync upgrade
    valid_commands = set([
        'upgrade', 'revision',
        'version', 'stamp', 'create_schema',
        'online_data_migrations',
    ])
    if not set(sys.argv) & valid_commands:
        sys.argv.append('upgrade')

    service.prepare_command(sys.argv)
    CONF.command.func()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

from oslo_config import cfg
from oslo_log import log
from oslo_messaging import exceptions as oslo_msg_exc
from oslo_utils import excutils
from oslo_versionedobjects import exception as oslo_vo_exc

from ironic.common import exception
from ironic.common.i18n import _
from ironic.objects import allocation as allocation_objects
from ironic.objects import chassis as chassis_objects
from ironic.objects import deploy_template as deploy_template_objects
from ironic.objects import fields
from ironic.objects import node as node_objects
from ironic.objects import notification
from ironic.objects import port as port_objects
from ironic.objects import portgroup as portgroup_objects
from ironic.objects import volume_connector as volume_connector_objects
from ironic.objects import volume_target as volume_target_objects

LOG = log.getLogger(__name__)
CONF = cfg.CONF


CRUD_NOTIFY_OBJ = {
    'allocation': (allocation_objects.AllocationCRUDNotification,
                   allocation_objects.AllocationCRUDPayload),
    'chassis': (chassis_objects.ChassisCRUDNotification,
                chassis_objects.ChassisCRUDPayload),
    'deploytemplate': (deploy_template_objects.DeployTemplateCRUDNotification,
                       deploy_template_objects.DeployTemplateCRUDPayload),
    'node': (node_objects.NodeCRUDNotification,
             node_objects.NodeCRUDPayload),
    'port': (port_objects.PortCRUDNotification,
             port_objects.PortCRUDPayload),
    'portgroup': (portgroup_objects.PortgroupCRUDNotification,
                  portgroup_objects.PortgroupCRUDPayload),
    'volumeconnector':
        (volume_connector_objects.VolumeConnectorCRUDNotification,
         volume_connector_objects.VolumeConnectorCRUDPayload),
    'volumetarget':
        (volume_target_objects.VolumeTargetCRUDNotification,
         volume_target_objects.VolumeTargetCRUDPayload),
}


def _emit_api_notification(context, obj, action, level, status, **kwargs):
    """Helper for emitting API notifications.

    :param context: request context.
    :param obj: resource rpc object.
    :param action: Action string to go in the EventType.
    :param level: Notification level. One of
                  `ironic.objects.fields.NotificationLevel.ALL`
    :param status: Status to go in the EventType. One of
                   `ironic.objects.fields.NotificationStatus.ALL`
    :param kwargs: kwargs to use when creating the notification payload.
    """
    resource = obj.__class__.__name__.lower()
    extra_args = kwargs
    try:
        try:
            if action == 'maintenance_set':
                notification_method = node_objects.NodeMaintenanceNotification
                payload_method = node_objects.NodePayload
            elif resource not in CRUD_NOTIFY_OBJ:
                notification_name = payload_name = _("is not defined")
                raise KeyError(_("Unsupported resource: %s") % resource)
            else:
                notification_method, payload_method = CRUD_NOTIFY_OBJ[resource]

            notification_name = notification_method.__name__
            payload_name = payload_method.__name__
        finally:
            # Prepare our exception message just in case
            exception_values = {"resource": resource,
                                "uuid": obj.uuid,
                                "action": action,
                                "status": status,
                                "level": level,
                                "notification_method": notification_name,
                                "payload_method": payload_name}
            exception_message = (_("Failed to send baremetal.%(resource)s."
                                   "%(action)s.%(status)s notification for "
                                   "%(resource)s %(uuid)s with level "
                                   "%(level)s, notification method "
                                   "%(notification_method)s, payload method "
                                   "%(payload_method)s, error %(error)s"))

        payload = payload_method(obj, **extra_args)
        if resource == 'node':
            notification.mask_secrets(payload)
        notification_method(
            publisher=notification.NotificationPublisher(
                service='ironic-api', host=CONF.host),
            event_type=notification.EventType(
                object=resource, action=action, status=status),
            level=level,
            payload=payload).emit(context)
    except (exception.NotificationSchemaObjectError,
            exception.NotificationSchemaKeyError,
            exception.NotificationPayloadError,
            oslo_msg_exc.MessageDeliveryFailure,
            oslo_vo_exc.VersionedObjectsException) as e:
        exception_values['error'] = e
        LOG.warning(exception_message, exception_values)
    except Exception as e:
        exception_values['error'] = e
        LOG.exception(exception_message, exception_values)


def emit_start_notification(context, obj, action, **kwargs):
    """Helper for emitting API 'start' notifications.

    :param context: request context.
    :param obj: resource rpc object.
    :param action: Action string to go in the EventType.
    :param kwargs: kwargs to use when creating the notification payload.
    """
    _emit_api_notification(context, obj, action,
                           fields.NotificationLevel.INFO,
                           fields.NotificationStatus.START,
                           **kwargs)


@contextlib.contextmanager
def handle_error_notification(context, obj, action, **kwargs):
    """Context manager to handle any error notifications.

    :param context: request context.
    :param obj: resource rpc object.
    :param action: Action string to go in the EventType.
    :param kwargs: kwargs to use when creating the notification payload.
    """
    try:
        yield
    except Exception:
        with excutils.save_and_reraise_exception():
            _emit_api_notification(context, obj, action,
                                   fields.NotificationLevel.ERROR,
                                   fields.NotificationStatus.ERROR,
                                   **kwargs)


def emit_end_notification(context, obj, action, **kwargs):
    """Helper for emitting API 'end' notifications.

    :param context: request context.
    :param obj: resource rpc object.
    :param action: Action string to go in the EventType.
    :param kwargs: kwargs to use when creating the notification payload.
    """
    _emit_api_notification(context, obj, action,
                           fields.NotificationLevel.INFO,
                           fields.NotificationStatus.END,
                           **kwargs)



# Copyright (c) 2015 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from ironic.common import release_mappings

CONF = cfg.CONF

# This is the version 1 API
BASE_VERSION = 1

# Here goes a short log of changes in every version.
# Refer to doc/source/contributor/webapi-version-history.rst for a detailed
# explanation of what each version contains.
#
# v1.0: corresponds to Juno API, not supported since Kilo
# v1.1: API at the point in time when versioning support was added,
# covers the following commits from Kilo cycle:
#   827db7fe: Add Node.maintenance_reason
#   68eed82b: Add API endpoint to set/unset the node maintenance mode
#   bc973889: Add sync and async support for passthru methods
#   e03f443b: Vendor endpoints to support different HTTP methods
#   e69e5309: Make vendor methods discoverable via the Ironic API
#   edf532db: Add logic to store the config drive passed by Nova
# v1.2: Renamed NOSTATE ("None") to AVAILABLE ("available")
# v1.3: Add node.driver_internal_info
# v1.4: Add MANAGEABLE state
# v1.5: Add logical node names
# v1.6: Add INSPECT* states
# v1.7: Add node.clean_step
# v1.8: Add ability to return a subset of resource fields
# v1.9: Add ability to filter nodes by provision state
# v1.10: Logical node names support RFC 3986 unreserved characters
# v1.11: Nodes appear in ENROLL state by default
# v1.12: Add support for RAID
# v1.13: Add 'abort' verb to CLEANWAIT
# v1.14: Make the following endpoints discoverable via API:
#        1. '/v1/nodes/<uuid>/states'
#        2. '/v1/drivers/<driver-name>/properties'
# v1.15: Add ability to do manual cleaning of nodes
# v1.16: Add ability to filter nodes by driver.
# v1.17: Add 'adopt' verb for ADOPTING active nodes.
# v1.18: Add port.internal_info.
# v1.19: Add port.local_link_connection and port.pxe_enabled.
# v1.20: Add node.network_interface
# v1.21: Add node.resource_class
# v1.22: Ramdisk lookup and heartbeat endpoints.
# v1.23: Add portgroup support.
# v1.24: Add subcontrollers: node.portgroup, portgroup.ports.
#        Add port.portgroup_uuid field.
# v1.25: Add possibility to unset chassis_uuid from node.
# v1.26: Add portgroup.mode and portgroup.properties.
# v1.27: Add soft reboot, soft power off and timeout.
# v1.28: Add vifs subcontroller to node
# v1.29: Add inject nmi.
# v1.30: Add dynamic driver interactions.
# v1.31: Add dynamic interfaces fields to node.
# v1.32: Add volume support.
# v1.33: Add node storage interface
# v1.34: Add physical network field to port.
# v1.35: Add ability to provide configdrive when rebuilding node.
# v1.36: Add Ironic Python Agent version support.
# v1.37: Add node traits.
# v1.38: Add rescue and unrescue provision states
# v1.39: Add inspect wait provision state.
# v1.40: Add bios.properties.
#        Add bios_interface to the node object.
# v1.41: Add inspection abort support.
# v1.42: Expose fault field to node.
# v1.43: Add detail=True flag to all API endpoints
# v1.44: Add node deploy_step field
# v1.45: reset_interfaces parameter to node's PATCH
# v1.46: Add conductor_group to the node object.
# v1.47: Add automated_clean to the node object.
# v1.48: Add protected to the node object.
# v1.49: Add conductor to the node object and /v1/conductors.
# v1.50: Add owner to the node object.
# v1.51: Add description to the node object.
# v1.52: Add allocation API.
# v1.53: Add support for Smart NIC port
# v1.54: Add events support.
# v1.55: Add deploy templates API.
# v1.56: Add support for building configdrives.
# v1.57: Add support for updating an exisiting allocation.
# v1.58: Add support for backfilling allocations.
# v1.59: Add support vendor data in configdrives.
# v1.60: Add owner to the allocation object.
# v1.61: Add retired and retired_reason to the node object.
# v1.62: Add agent_token support for agent communication.
# v1.63: Add support for indicators
# v1.64: Add network_type to port.local_link_connection
# v1.65: Add lessee to the node object.
# v1.66: Add support for node network_data field.
# v1.67: Add support for port_uuid/portgroup_uuid in node vif_attach
# v1.68: Add agent_verify_ca to heartbeat.
# v1.69: Add deploy_steps to provisioning
# v1.70: Add disable_ramdisk to manual cleaning.
# v1.71: Add signifier for Scope based roles.
# v1.72: Add agent_status and agent_status_message to /v1/heartbeat
# v1.73: Add support for deploy and undeploy verbs
# v1.74: Add bios registry to /v1/nodes/{node}/bios/{setting}
# v1.75: Add boot_mode, secure_boot fields to node object.
# v1.76: Add support for changing boot_mode and secure_boot state
# v1.77: Add fields selector to drivers list and driver detail.
# v1.78: Add node history endpoint

MINOR_0_JUNO = 0
MINOR_1_INITIAL_VERSION = 1
MINOR_2_AVAILABLE_STATE = 2
MINOR_3_DRIVER_INTERNAL_INFO = 3
MINOR_4_MANAGEABLE_STATE = 4
MINOR_5_NODE_NAME = 5
MINOR_6_INSPECT_STATE = 6
MINOR_7_NODE_CLEAN = 7
MINOR_8_FETCHING_SUBSET_OF_FIELDS = 8
MINOR_9_PROVISION_STATE_FILTER = 9
MINOR_10_UNRESTRICTED_NODE_NAME = 10
MINOR_11_ENROLL_STATE = 11
MINOR_12_RAID_CONFIG = 12
MINOR_13_ABORT_VERB = 13
MINOR_14_LINKS_NODESTATES_DRIVERPROPERTIES = 14
MINOR_15_MANUAL_CLEAN = 15
MINOR_16_DRIVER_FILTER = 16
MINOR_17_ADOPT_VERB = 17
MINOR_18_PORT_INTERNAL_INFO = 18
MINOR_19_PORT_ADVANCED_NET_FIELDS = 19
MINOR_20_NETWORK_INTERFACE = 20
MINOR_21_RESOURCE_CLASS = 21
MINOR_22_LOOKUP_HEARTBEAT = 22
MINOR_23_PORTGROUPS = 23
MINOR_24_PORTGROUPS_SUBCONTROLLERS = 24
MINOR_25_UNSET_CHASSIS_UUID = 25
MINOR_26_PORTGROUP_MODE_PROPERTIES = 26
MINOR_27_SOFT_POWER_OFF = 27
MINOR_28_VIFS_SUBCONTROLLER = 28
MINOR_29_INJECT_NMI = 29
MINOR_30_DYNAMIC_DRIVERS = 30
MINOR_31_DYNAMIC_INTERFACES = 31
MINOR_32_VOLUME = 32
MINOR_33_STORAGE_INTERFACE = 33
MINOR_34_PORT_PHYSICAL_NETWORK = 34
MINOR_35_REBUILD_CONFIG_DRIVE = 35
MINOR_36_AGENT_VERSION_HEARTBEAT = 36
MINOR_37_NODE_TRAITS = 37
MINOR_38_RESCUE_INTERFACE = 38
MINOR_39_INSPECT_WAIT = 39
MINOR_40_BIOS_INTERFACE = 40
MINOR_41_INSPECTION_ABORT = 41
MINOR_42_FAULT = 42
MINOR_43_ENABLE_DETAIL_QUERY = 43
MINOR_44_NODE_DEPLOY_STEP = 44
MINOR_45_RESET_INTERFACES = 45
MINOR_46_NODE_CONDUCTOR_GROUP = 46
MINOR_47_NODE_AUTOMATED_CLEAN = 47
MINOR_48_NODE_PROTECTED = 48
MINOR_49_CONDUCTORS = 49
MINOR_50_NODE_OWNER = 50
MINOR_51_NODE_DESCRIPTION = 51
MINOR_52_ALLOCATION = 52
MINOR_53_PORT_SMARTNIC = 53
MINOR_54_EVENTS = 54
MINOR_55_DEPLOY_TEMPLATES = 55
MINOR_56_BUILD_CONFIGDRIVE = 56
MINOR_57_ALLOCATION_UPDATE = 57
MINOR_58_ALLOCATION_BACKFILL = 58
MINOR_59_CONFIGDRIVE_VENDOR_DATA = 59
MINOR_60_ALLOCATION_OWNER = 60
MINOR_61_NODE_RETIRED = 61
MINOR_62_AGENT_TOKEN = 62
MINOR_63_INDICATORS = 63
MINOR_64_LOCAL_LINK_CONNECTION_NETWORK_TYPE = 64
MINOR_65_NODE_LESSEE = 65
MINOR_66_NODE_NETWORK_DATA = 66
MINOR_67_NODE_VIF_ATTACH_PORT = 67
MINOR_68_HEARTBEAT_VERIFY_CA = 68
MINOR_69_DEPLOY_STEPS = 69
MINOR_70_CLEAN_DISABLE_RAMDISK = 70
MINOR_71_RBAC_SCOPES = 71
MINOR_72_HEARTBEAT_STATUS = 72
MINOR_73_DEPLOY_UNDEPLOY_VERBS = 73
MINOR_74_BIOS_REGISTRY = 74
MINOR_75_NODE_BOOT_MODE = 75
MINOR_76_NODE_CHANGE_BOOT_MODE = 76
MINOR_77_DRIVER_FIELDS_SELECTOR = 77
MINOR_78_NODE_HISTORY = 78

# When adding another version, update:
# - MINOR_MAX_VERSION
# - doc/source/contributor/webapi-version-history.rst with a detailed
#   explanation of what changed in the new version
# - common/release_mappings.py, RELEASE_MAPPING['master']['api']

MINOR_MAX_VERSION = MINOR_78_NODE_HISTORY

# String representations of the minor and maximum versions
_MIN_VERSION_STRING = '{}.{}'.format(BASE_VERSION, MINOR_1_INITIAL_VERSION)
_MAX_VERSION_STRING = '{}.{}'.format(BASE_VERSION, MINOR_MAX_VERSION)


def min_version_string():
    """Returns the minimum supported API version (as a string)"""
    return _MIN_VERSION_STRING


def max_version_string():
    """Returns the maximum supported API version (as a string).

    If the service is pinned, the maximum API version is the pinned
    version. Otherwise, it is the maximum supported API version.
    """
    release_ver = release_mappings.RELEASE_MAPPING.get(
        CONF.pin_release_version)
    if release_ver:
        return release_ver['api']
    else:
        return _MAX_VERSION_STRING



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import os

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils
import pbr.version


CONF = cfg.CONF
DOMAIN = 'ironic'
VERSION = pbr.version.VersionInfo(DOMAIN).release_string()


# find and parse callback config file
def parse_callback_config():
    basename = os.path.splitext(__file__)[0]
    config = configparser.ConfigParser()
    callback_config = {'ironic_config': None,
                       'ironic_log_file': None,
                       'use_journal': True,
                       'use_syslog': False}
    try:
        config.read_file(open(basename + ".ini"))
        if config.has_option('ironic', 'config_file'):
            callback_config['ironic_config'] = config.get(
                'ironic', 'config_file')
        if config.has_option('ironic', 'log_file'):
            callback_config['ironic_log_file'] = config.get(
                'ironic', 'log_file')
        if config.has_option('ironic', 'use_journal'):
            callback_config['use_journal'] = strutils.bool_from_string(
                config.get('ironic', 'use_journal'))
        if config.has_option('ironic', 'use_syslog'):
            callback_config['use_syslog'] = strutils.bool_from_string(
                config.get('ironic', 'use_syslog'))
    except Exception:
        pass
    return callback_config


def setup_log():

    logging.register_options(CONF)

    conf_kwargs = dict(args=[], project=DOMAIN, version=VERSION)
    callback_config = parse_callback_config()

    if callback_config['ironic_config']:
        conf_kwargs['default_config_files'] = [
            callback_config['ironic_config']]
    CONF(**conf_kwargs)

    if callback_config['use_journal']:
        CONF.set_override('use_journal', True)
    if callback_config['use_syslog']:
        CONF.set_override('use_syslog', True)
    if callback_config['ironic_log_file']:
        CONF.set_override("log_file", callback_config['ironic_log_file'])

    logging.setup(CONF, DOMAIN)


class CallbackModule(object):

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'notification'
    CALLBACK_NAME = 'ironic_log'
    CALLBACK_NEEDS_WHITELIST = True

    def __init__(self, display=None):
        setup_log()
        self.log = logging.getLogger(__name__)
        self.node = None
        self._plugin_options = {}

    # NOTE(pas-ha) this method is required for Ansible>=2.4
    # TODO(pas-ha) rewrite to support defining callback plugin options
    # in ansible.cfg after we require Ansible >=2.4
    def set_options(self, option=None, option_value=None):
        if option:
            if option_value:
                self._plugin_options[option] = option_value
            else:
                self._plugin_options = option

    def runner_msg_dict(self, result):
        self.node = result._host.get_name()
        name = result._task.get_name()
        res = str(result._result)
        return dict(node=self.node, name=name, res=res)

    def v2_playbook_on_task_start(self, task, is_conditional):
        # NOTE(pas-ha) I do not know (yet) how to obtain a ref to host
        # until first task is processed
        node = self.node or "Node"
        name = task.get_name()
        if name == 'setup':
            self.log.debug("Processing task %(name)s.", dict(name=name))
        else:
            self.log.debug("Processing task %(name)s on node %(node)s.",
                           dict(name=name, node=node))

    def v2_runner_on_failed(self, result, *args, **kwargs):
        self.log.error(
            "Ansible task %(name)s failed on node %(node)s: %(res)s",
            self.runner_msg_dict(result))

    def v2_runner_on_ok(self, result):
        msg_dict = self.runner_msg_dict(result)
        if msg_dict['name'] == 'setup':
            self.log.info("Ansible task 'setup' complete on node %(node)s",
                          msg_dict)
        else:
            self.log.info("Ansible task %(name)s complete on node %(node)s: "
                          "%(res)s", msg_dict)

    def v2_runner_on_unreachable(self, result):
        self.log.error(
            "Node %(node)s was unreachable for Ansible task %(name)s: %(res)s",
            self.runner_msg_dict(result))

    def v2_runner_on_async_poll(self, result):
        self.log.debug("Polled ansible task %(name)s for complete "
                       "on node %(node)s: %(res)s",
                       self.runner_msg_dict(result))

    def v2_runner_on_async_ok(self, result):
        self.log.info("Async Ansible task %(name)s complete on node %(node)s: "
                      "%(res)s", self.runner_msg_dict(result))

    def v2_runner_on_async_failed(self, result):
        self.log.error("Async Ansible task %(name)s failed on node %(node)s: "
                       "%(res)s", self.runner_msg_dict(result))

    def v2_runner_on_skipped(self, result):
        self.log.debug(
            "Ansible task %(name)s skipped on node %(node)s: %(res)s",
            self.runner_msg_dict(result))



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
:mod:`ironic.tests.unit` -- ironic unit tests
=====================================================

.. automodule:: ironic.tests.unit
   :platform: Unix
"""

# TODO(tenbrae): move eventlet imports to ironic.__init__ once we move to PBR

import eventlet
from oslo_config import cfg
from oslo_log import log

from ironic import objects

eventlet.monkey_patch(os=False)

log.register_options(cfg.CONF)
log.setup(cfg.CONF, 'ironic')

# NOTE(comstud): Make sure we have all of the objects loaded. We do this
# at module import time, because we may be using mock decorators in our
# tests that run at import time.
objects.register_all()

# NOTE(dtantsur): this module creates mocks which may be used at random points
# of time, so it must be imported as early as possible.
from ironic.tests.unit.drivers import third_party_driver_mocks   # noqa



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client

from ironic_lib import metrics_utils
from pecan import rest

from ironic import api
from ironic.api.controllers import link
from ironic.api.controllers.v1 import utils as api_utils
from ironic.api import method
from ironic.common import args
from ironic.common import exception
from ironic.common.i18n import _
from ironic.drivers import base as driver_base


METRICS = metrics_utils.get_metrics_logger(__name__)

# Property information for drivers:
#   key = driver name;
#   value = dictionary of properties of that driver:
#             key = property name.
#             value = description of the property.
# NOTE(rloo). This is cached for the lifetime of the API service. If one or
# more conductor services are restarted with new driver versions, the API
# service should be restarted.
_DRIVER_PROPERTIES = {}

# Vendor information for drivers:
#   key = driver name;
#   value = dictionary of vendor methods of that driver:
#             key = method name.
#             value = dictionary with the metadata of that method.
# NOTE(lucasagomes). This is cached for the lifetime of the API
# service. If one or more conductor services are restarted with new driver
# versions, the API service should be restarted.
_VENDOR_METHODS = {}

# RAID (logical disk) configuration information for drivers:
#   key = driver name;
#   value = dictionary of RAID configuration information of that driver:
#             key = property name.
#             value = description of the property
# NOTE(rloo). This is cached for the lifetime of the API service. If one or
# more conductor services are restarted with new driver versions, the API
# service should be restarted.
_RAID_PROPERTIES = {}


def hide_fields_in_newer_versions(driver):
    """This method hides fields that were added in newer API versions.

    Certain fields were introduced at certain API versions.
    These fields are only made available when the request's API version
    matches or exceeds the versions when these fields were introduced.
    """
    if not api_utils.allow_storage_interface():
        driver.pop('default_storage_interface', None)
        driver.pop('enabled_storage_interfaces', None)

    if not api_utils.allow_rescue_interface():
        driver.pop('default_rescue_interface', None)
        driver.pop('enabled_rescue_interfaces', None)

    if not api_utils.allow_bios_interface():
        driver.pop('default_bios_interface', None)
        driver.pop('enabled_bios_interfaces', None)


def convert_with_links(name, hosts, detail=False, interface_info=None,
                       fields=None, sanitize=True):
    """Convert driver/hardware type info to a dict.

    :param name: name of a hardware type.
    :param hosts: list of conductor hostnames driver is active on.
    :param detail: boolean, whether to include detailed info, such as
                    the 'type' field and default/enabled interfaces fields.
    :param interface_info: optional list of dicts of hardware interface
                            info.
    :param fields: list of fields to preserve, or ``None`` to preserve default
    :param sanitize: boolean,  sanitize driver
    :returns: dict representing the driver object.
    """
    driver = {
        'name': name,
        'hosts': hosts,
        'links': [
            link.make_link('self',
                           api.request.public_url,
                           'drivers', name),
            link.make_link('bookmark',
                           api.request.public_url,
                           'drivers', name,
                           bookmark=True)
        ]
    }
    if api_utils.allow_links_node_states_and_driver_properties():
        driver['properties'] = [
            link.make_link('self',
                           api.request.public_url,
                           'drivers', name + "/properties"),
            link.make_link('bookmark',
                           api.request.public_url,
                           'drivers', name + "/properties",
                           bookmark=True)
        ]

    if api_utils.allow_dynamic_drivers():
        # NOTE(dtantsur): only dynamic drivers (based on hardware types)
        # are supported starting with the Rocky release.
        driver['type'] = 'dynamic'
        if detail:
            if interface_info is None:
                # TODO(jroll) objectify this
                interface_info = (api.request.dbapi
                                  .list_hardware_type_interfaces([name]))
            for iface_type in driver_base.ALL_INTERFACES:
                default = None
                enabled = set()
                for iface in interface_info:
                    if iface['interface_type'] == iface_type:
                        iface_name = iface['interface_name']
                        enabled.add(iface_name)
                        # NOTE(jroll) this assumes the default is the same
                        # on all conductors
                        if iface['default']:
                            default = iface_name

                default_key = 'default_%s_interface' % iface_type
                enabled_key = 'enabled_%s_interfaces' % iface_type
                driver[default_key] = default
                driver[enabled_key] = list(enabled)

    hide_fields_in_newer_versions(driver)

    if not sanitize:
        return driver

    driver_sanitize(driver, fields)

    return driver


def driver_sanitize(driver, fields=None):
    if fields is not None:
        api_utils.sanitize_dict(driver, fields)
        api_utils.check_for_invalid_fields(fields, driver)


def _check_allow_driver_fields(fields):
    if (fields is not None and api.request.version.minor
            < api.controllers.v1.versions.MINOR_77_DRIVER_FIELDS_SELECTOR):
        raise exception.NotAcceptable()


def list_convert_with_links(hardware_types, detail=False, fields=None):
    """Convert drivers and hardware types to an API-serializable object.

    :param hardware_types: dict mapping hardware type names to conductor
                            hostnames.
    :param detail: boolean, whether to include detailed info, such as
                    the 'type' field and default/enabled interfaces fields.
    :param fields: list of fields to preserve, or ``None`` to preserve default
    :returns: an API-serializable driver collection object.
    """
    drivers = []
    collection = {
        'drivers': drivers
    }

    # NOTE(jroll) we return hardware types in all API versions,
    # but restrict type/default/enabled fields to 1.30.
    # This is checked in Driver.convert_with_links(), however also
    # checking here can save us a DB query.
    if api_utils.allow_dynamic_drivers() and detail:
        iface_info = api.request.dbapi.list_hardware_type_interfaces(
            list(hardware_types))
    else:
        iface_info = []

    for htname in hardware_types:
        interface_info = [i for i in iface_info
                          if i['hardware_type'] == htname]
        drivers.append(
            convert_with_links(htname,
                               list(hardware_types[htname]),
                               detail=detail,
                               interface_info=interface_info,
                               fields=fields))
    return collection


class DriverPassthruController(rest.RestController):
    """REST controller for driver passthru.

    This controller allow vendors to expose cross-node functionality in the
    Ironic API. Ironic will merely relay the message from here to the specified
    driver, no introspection will be made in the message body.
    """

    _custom_actions = {
        'methods': ['GET']
    }

    @METRICS.timer('DriverPassthruController.methods')
    @method.expose()
    @args.validate(driver_name=args.string)
    def methods(self, driver_name):
        """Retrieve information about vendor methods of the given driver.

        :param driver_name: name of the driver.
        :returns: dictionary with <vendor method name>:<method metadata>
                  entries.
        :raises DriverNotFound: if the driver name is invalid or the
                 driver cannot be loaded.
        """
        api_utils.check_policy('baremetal:driver:vendor_passthru')

        if driver_name not in _VENDOR_METHODS:
            topic = api.request.rpcapi.get_topic_for_driver(driver_name)
            ret = api.request.rpcapi.get_driver_vendor_passthru_methods(
                api.request.context, driver_name, topic=topic)
            _VENDOR_METHODS[driver_name] = ret

        return _VENDOR_METHODS[driver_name]

    @METRICS.timer('DriverPassthruController._default')
    @method.expose()
    @method.body('data')
    @args.validate(driver_name=args.string, method=args.string)
    def _default(self, driver_name, method, data=None):
        """Call a driver API extension.

        :param driver_name: name of the driver to call.
        :param method: name of the method, to be passed to the vendor
                       implementation.
        :param data: body of data to supply to the specified method.
        """

        api_utils.check_policy('baremetal:driver:vendor_passthru')

        topic = api.request.rpcapi.get_topic_for_driver(driver_name)
        resp = api_utils.vendor_passthru(driver_name, method, topic,
                                         data=data, driver_passthru=True)
        api.response.status_code = resp.status_code
        return resp.obj


class DriverRaidController(rest.RestController):

    _custom_actions = {
        'logical_disk_properties': ['GET']
    }

    @METRICS.timer('DriverRaidController.logical_disk_properties')
    @method.expose()
    @args.validate(driver_name=args.string)
    def logical_disk_properties(self, driver_name):
        """Returns the logical disk properties for the driver.

        .. parameters:: ../../api-ref/source/parameters.yaml

           :driver_name: Name of the driver.

        .. return::

           Success:
           A dictionary containing the properties that can be mentioned

           Failure:
           :UnsupportedDriverExtension: If the driver doesn't support RAID
                                        configuration.
           :NotAcceptable: If requested version of the API is less than 1.12.
           :DriverNotFound: If driver is not loaded on any of the conductors.
        """
        api_utils.check_policy(
            'baremetal:driver:get_raid_logical_disk_properties')

        if not api_utils.allow_raid_config():
            raise exception.NotAcceptable()

        if driver_name not in _RAID_PROPERTIES:
            topic = api.request.rpcapi.get_topic_for_driver(driver_name)
            try:
                info = api.request.rpcapi.get_raid_logical_disk_properties(
                    api.request.context, driver_name, topic=topic)
            except exception.UnsupportedDriverExtension as e:
                # Change error code as 404 seems appropriate because RAID is a
                # standard interface and all drivers might not have it.
                e.code = http_client.NOT_FOUND
                raise

            _RAID_PROPERTIES[driver_name] = info
        return _RAID_PROPERTIES[driver_name]


class DriversController(rest.RestController):
    """REST controller for Drivers."""

    vendor_passthru = DriverPassthruController()

    raid = DriverRaidController()
    """Expose RAID as a sub-element of drivers"""

    _custom_actions = {
        'properties': ['GET'],
    }

    @METRICS.timer('DriversController.get_all')
    @method.expose()
    @args.validate(type=args.string, detail=args.boolean,
                   fields=args.string_list)
    def get_all(self, type=None, detail=None, fields=None):
        """Retrieve a list of drivers."""
        # FIXME(tenbrae): formatting of the auto-generated REST API docs
        #              will break from a single-line doc string.
        #              This is a result of a bug in sphinxcontrib-pecanwsme
        # https://github.com/dreamhost/sphinxcontrib-pecanwsme/issues/8
        if fields and detail:
            raise exception.InvalidParameterValue(
                "Can not specify ?detail=True and fields in the same request.")

        api_utils.check_policy('baremetal:driver:get')
        api_utils.check_allow_driver_detail(detail)
        api_utils.check_allow_filter_driver_type(type)
        _check_allow_driver_fields(fields)
        if type not in (None, 'classic', 'dynamic'):
            raise exception.Invalid(_(
                '"type" filter must be one of "classic" or "dynamic", '
                'if specified.'))

        if type is None or type == 'dynamic':
            hw_type_dict = api.request.dbapi.get_active_hardware_type_dict()
        else:
            # NOTE(dtantsur): we don't support classic drivers starting with
            # the Rocky release.
            hw_type_dict = {}
        return list_convert_with_links(hw_type_dict, detail=detail,
                                       fields=fields)

    @METRICS.timer('DriversController.get_one')
    @method.expose()
    @args.validate(driver_name=args.string, fields=args.string_list)
    def get_one(self, driver_name, fields=None):
        """Retrieve a single driver."""
        # NOTE(russell_h): There is no way to make this more efficient than
        # retrieving a list of drivers using the current sqlalchemy schema, but
        # this path must be exposed for Pecan to route any paths we might
        # choose to expose below it.
        api_utils.check_policy('baremetal:driver:get')

        _check_allow_driver_fields(fields)

        hw_type_dict = api.request.dbapi.get_active_hardware_type_dict()
        for name, hosts in hw_type_dict.items():
            if name == driver_name:
                return convert_with_links(name, list(hosts),
                                          detail=True, fields=fields)

        raise exception.DriverNotFound(driver_name=driver_name)

    @METRICS.timer('DriversController.properties')
    @method.expose()
    @args.validate(driver_name=args.string)
    def properties(self, driver_name):
        """Retrieve property information of the given driver.

        :param driver_name: name of the driver.
        :returns: dictionary with <property name>:<property description>
                  entries.
        :raises DriverNotFound (HTTP 404): if the driver name is invalid or
                 the driver cannot be loaded.
        """
        api_utils.check_policy('baremetal:driver:get_properties')

        if driver_name not in _DRIVER_PROPERTIES:
            topic = api.request.rpcapi.get_topic_for_driver(driver_name)
            properties = api.request.rpcapi.get_driver_properties(
                api.request.context, driver_name, topic=topic)
            _DRIVER_PROPERTIES[driver_name] = properties

        return _DRIVER_PROPERTIES[driver_name]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from ironic.common import exception
from ironic.tests.unit.db import base
from ironic.tests.unit.db import utils as db_utils


class DBNodeHistoryTestCase(base.DbTestCase):

    def setUp(self):
        super(DBNodeHistoryTestCase, self).setUp()
        self.node = db_utils.create_test_node()
        self.history = db_utils.create_test_history(
            id=0, node_id=self.node.id, conductor='test-conductor',
            user='fake-user', event='Something bad happened but fear not')

    def test_destroy_node_history_by_uuid(self):
        self.dbapi.destroy_node_history_by_uuid(self.history.uuid)
        self.assertRaises(exception.NodeHistoryNotFound,
                          self.dbapi.get_node_history_by_id,
                          self.history.id)
        self.assertRaises(exception.NodeHistoryNotFound,
                          self.dbapi.get_node_history_by_uuid,
                          self.history.uuid)

    def test_get_history_by_id(self):
        res = self.dbapi.get_node_history_by_id(self.history.id)
        self.assertEqual(self.history.conductor, res.conductor)
        self.assertEqual(self.history.user, res.user)
        self.assertEqual(self.history.event, res.event)

    def test_get_history_by_id_not_found(self):
        self.assertRaises(exception.NodeHistoryNotFound,
                          self.dbapi.get_node_history_by_id, -1)

    def test_get_history_by_uuid(self):
        res = self.dbapi.get_node_history_by_uuid(self.history.uuid)
        self.assertEqual(self.history.id, res.id)

    def test_get_history_by_uuid_not_found(self):
        self.assertRaises(exception.NodeHistoryNotFound,
                          self.dbapi.get_node_history_by_uuid,
                          'wrong-uuid')

    def _prepare_history_entries(self):
        uuids = [str(self.history.uuid)]
        for i in range(1, 6):
            history = db_utils.create_test_history(
                id=i, uuid=uuidutils.generate_uuid(),
                node_id=self.node.id,
                conductor='test-conductor', user='fake-user',
                event='Something bad happened but fear not %s' % i,
                severity='ERROR', event_type='test')
            uuids.append(str(history.uuid))
        return uuids

    def test_get_node_history_list(self):
        uuids = self._prepare_history_entries()
        res = self.dbapi.get_node_history_list()
        res_uuids = [r.uuid for r in res]
        self.assertCountEqual(uuids, res_uuids)

    def test_get_node_history_list_sorted(self):
        self._prepare_history_entries()

        res = self.dbapi.get_node_history_list(sort_key='created_at',
                                               sort_dir='desc')
        expected = sorted(res, key=lambda r: r.created_at, reverse=True)
        self.assertEqual(res, expected)
        self.assertIn('fear not 5', res[0].event)

    def test_get_history_by_node_id_empty(self):
        self.assertEqual([], self.dbapi.get_node_history_by_node_id(10))

    def test_get_history_by_node_id(self):
        res = self.dbapi.get_node_history_by_node_id(self.node.id)
        self.assertEqual(self.history.uuid, res[0].uuid)
        self.assertEqual(self.history.user, res[0].user)
        self.assertEqual(self.history.conductor, res[0].conductor)
        self.assertEqual(self.history.event, res[0].event)
        self.assertEqual(self.history.event_type, res[0].event_type)
        self.assertEqual(self.history.severity, res[0].severity)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_concurrency import processutils
from oslo_config import cfg

from ironic.common import exception
from ironic.common import wsgi_service
from ironic.tests import base

CONF = cfg.CONF


class TestWSGIService(base.TestCase):
    @mock.patch.object(processutils, 'get_worker_count', lambda: 2)
    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_default(self, mock_server):
        service_name = "ironic_api"
        test_service = wsgi_service.WSGIService(service_name)
        self.assertEqual(2, test_service.workers)
        mock_server.assert_called_once_with(CONF, service_name,
                                            test_service.app,
                                            host='0.0.0.0',
                                            port=6385,
                                            use_ssl=False)

    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_correct_setting(self, mock_server):
        self.config(api_workers=8, group='api')
        test_service = wsgi_service.WSGIService("ironic_api")
        self.assertEqual(8, test_service.workers)

    @mock.patch.object(processutils, 'get_worker_count', lambda: 3)
    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_zero_setting(self, mock_server):
        self.config(api_workers=0, group='api')
        test_service = wsgi_service.WSGIService("ironic_api")
        self.assertEqual(3, test_service.workers)

    @mock.patch.object(processutils, 'get_worker_count', lambda: 42)
    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_default_limit(self, mock_server):
        self.config(api_workers=0, group='api')
        test_service = wsgi_service.WSGIService("ironic_api")
        self.assertEqual(4, test_service.workers)

    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_negative_setting(self, mock_server):
        self.config(api_workers=-2, group='api')
        self.assertRaises(exception.ConfigInvalid,
                          wsgi_service.WSGIService,
                          'ironic_api')
        self.assertFalse(mock_server.called)

    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_wsgi_service_with_ssl_enabled(self, mock_server):
        self.config(enable_ssl_api=True, group='api')
        service_name = 'ironic_api'
        srv = wsgi_service.WSGIService('ironic_api', CONF.api.enable_ssl_api)
        mock_server.assert_called_once_with(CONF, service_name,
                                            srv.app,
                                            host='0.0.0.0',
                                            port=6385,
                                            use_ssl=True)



#!/usr/bin/env python3
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import argparse
import os.path
import string
import sys

import jinja2
import libvirt

templatedir = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                           'templates')


CONSOLE_LOG = """
    <serial type='file'>
      <source path='%(console_log)s'/>
      <target port='0'/>
      <alias name='serial0'/>
    </serial>
    <serial type='pty'>
      <source path='/dev/pts/49'/>
      <target port='1'/>
      <alias name='serial1'/>
    </serial>
    <console type='file'>
      <source path='%(console_log)s'/>
      <target type='serial' port='0'/>
      <alias name='serial0'/>
    </console>
"""


CONSOLE_PTY = """
    <serial type='pty'>
      <target port='0'/>
    </serial>
    <console type='pty'>
      <target type='serial' port='0'/>
    </console>
"""


def main():
    parser = argparse.ArgumentParser(
        description="Configure a kvm virtual machine for the seed image.")
    parser.add_argument('--name', default='seed',
                        help='the name to give the machine in libvirt.')
    parser.add_argument('--image', action='append', default=[],
                        help='Use a custom image file (must be qcow2).')
    parser.add_argument('--engine', default='qemu',
                        help='The virtualization engine to use')
    parser.add_argument('--arch', default='i686',
                        help='The architecture to use')
    parser.add_argument('--memory', default='2097152',
                        help="Maximum memory for the VM in KB.")
    parser.add_argument('--cpus', default='1',
                        help="CPU count for the VM.")
    parser.add_argument('--bootdev', default='hd',
                        help="What boot device to use (hd/network).")
    parser.add_argument('--libvirt-nic-driver', default='virtio',
                        help='The libvirt network driver to use')
    parser.add_argument('--interface-count', default=1, type=int,
                        help='The number of interfaces to add to VM.'),
    parser.add_argument('--mac', default=None,
                        help='The mac for the first interface on the vm')
    parser.add_argument('--console-log',
                        help='File to log console')
    parser.add_argument('--emulator', default=None,
                        help='Path to emulator bin for vm template')
    parser.add_argument('--disk-format', default='qcow2',
                        help='Disk format to use.')
    parser.add_argument('--uefi-loader', default='',
                        help='The absolute path of the UEFI firmware blob.')
    parser.add_argument('--uefi-nvram', default='',
                        help=('The absolute path of the non-volatile memory '
                              'to store the UEFI variables. Should be used '
                              'only when --uefi-loader is also specified.'))
    args = parser.parse_args()

    env = jinja2.Environment(loader=jinja2.FileSystemLoader(templatedir))
    template = env.get_template('vm.xml')

    images = list(zip(args.image, string.ascii_lowercase))
    if not images or len(images) > 6:
        # 6 is an artificial limitation because of the way we generate PCI IDs
        sys.exit("Up to 6 images are required")

    params = {
        'name': args.name,
        'images': images,
        'engine': args.engine,
        'arch': args.arch,
        'memory': args.memory,
        'cpus': args.cpus,
        'bootdev': args.bootdev,
        'interface_count': args.interface_count,
        'mac': args.mac,
        'nicdriver': args.libvirt_nic_driver,
        'emulator': args.emulator,
        'disk_format': args.disk_format,
        'uefi_loader': args.uefi_loader,
        'uefi_nvram': args.uefi_nvram,
    }

    if args.emulator:
        params['emulator'] = args.emulator
    else:
        qemu_kvm_locations = ['/usr/bin/kvm',
                              '/usr/bin/qemu-kvm',
                              '/usr/libexec/qemu-kvm']
        for location in qemu_kvm_locations:
            if os.path.exists(location):
                params['emulator'] = location
                break
        else:
            raise RuntimeError("Unable to find location of kvm executable")

    if args.console_log:
        params['console'] = CONSOLE_LOG % {'console_log': args.console_log}
    else:
        params['console'] = CONSOLE_PTY
    libvirt_template = template.render(**params)
    conn = libvirt.open("qemu:///system")

    a = conn.defineXML(libvirt_template)
    print("Created machine %s with UUID %s" % (args.name, a.UUIDString()))


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add node tags

Revision ID: 48d6c242bb9b
Revises: 516faf1bb9b1
Create Date: 2015-10-08 10:07:33.779516

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '48d6c242bb9b'
down_revision = '516faf1bb9b1'


def upgrade():
    op.create_table(
        'node_tags',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('node_id', sa.Integer(), nullable=False,
                  autoincrement=False),
        sa.Column('tag', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['node_id'], ['nodes.id'], ),
        sa.PrimaryKeyConstraint('node_id', 'tag'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )
    op.create_index('node_tags_idx', 'node_tags', ['tag'], unique=False)
